#! /usr/bin/env python
"""Calculate gradients on a raster grid.

Gradient calculators for raster grids
+++++++++++++++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.raster_gradients.calc_grad_at_link
    ~landlab.grid.raster_gradients.calc_grad_at_active_link
    ~landlab.grid.raster_gradients.calc_grad_across_cell_faces
    ~landlab.grid.raster_gradients.calc_grad_across_cell_corners
    ~landlab.grid.raster_gradients.alculate_gradient_along_node_links

"""
import numpy as np

from landlab.core.utils import make_optional_arg_into_id_array, \
    radians_to_degrees

from landlab.grid import gradients
from landlab.grid.base import BAD_INDEX_VALUE, CLOSED_BOUNDARY
from landlab.utils.decorators import use_field_name_or_array
from collections import deque


@use_field_name_or_array('node')
def calc_grad_at_link(grid, node_values, out=None):
    """Calculate gradients in node_values at links.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    node_values : array_like or field name
        Values at nodes.
    out : ndarray, optional
        Buffer to hold result. If `None`, create a new array.

    Returns
    -------
    ndarray
        Gradients of the nodes values for each link.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((3, 3))
    >>> node_values = [0., 0., 0.,
    ...                1., 3., 1.,
    ...                2., 2., 2.]
    >>> grid.calc_grad_at_link(node_values)
    array([ 0.,  0.,  1.,  3.,  1.,  2., -2.,  1., -1.,  1.,  0.,  0.])

    >>> out = np.empty(grid.number_of_links, dtype=float)
    >>> rtn = grid.calc_grad_at_link(node_values, out=out)
    >>> rtn is out
    True
    >>> out
    array([ 0.,  0.,  1.,  3.,  1.,  2., -2.,  1., -1.,  1.,  0.,  0.])

    >>> grid = RasterModelGrid((3, 3), spacing=(1, 2))
    >>> grid.calc_grad_at_link(node_values)
    array([ 0.,  0.,  1.,  3.,  1.,  1., -1.,  1., -1.,  1.,  0.,  0.])
    >>> _ = grid.add_field('node', 'elevation', node_values)
    >>> grid.calc_grad_at_link('elevation')
    array([ 0.,  0.,  1.,  3.,  1.,  1., -1.,  1., -1.,  1.,  0.,  0.])

    LLCATS: LINF GRAD
    """
    grads = gradients.calc_diff_at_link(grid, node_values, out=out)
    grads /= grid.length_of_link[:grid.number_of_links]

#    n_vertical_links = (grid.shape[0] - 1) * grid.shape[1]
#    diffs[:n_vertical_links] /= grid.dy
#    diffs[n_vertical_links:] /= grid.dx

    return grads


@use_field_name_or_array('node')
def calc_grad_at_active_link(grid, node_values, out=None):
    """Calculate gradients over active links.

    .. deprecated:: 0.1
        Use :func:`calc_grad_across_cell_faces`
                or :func:`calc_grad_across_cell_corners` instead

    Calculates the gradient in quantity s at each active link in the grid.
    This is nearly identical to the method of the same name in ModelGrid,
    except that it uses a constant node spacing for link length to improve
    efficiency.

    Note that a negative gradient corresponds to a lower node in the
    direction of the link.

    Returns
    -------
    ndarray
        Gradients of the nodes values for each link.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(4, 5, 1.0)
    >>> u = [0., 1., 2., 3., 0.,
    ...      1., 2., 3., 2., 3.,
    ...      0., 1., 2., 1., 2.,
    ...      0., 0., 2., 2., 0.]
    >>> grad = grid.calc_grad_at_active_link(u)
    >>> grad # doctest: +NORMALIZE_WHITESPACE
    array([ 1.,  1., -1.,
            1.,  1., -1.,  1.,
           -1., -1., -1.,
            1.,  1., -1.,  1.,
           -1.,  0.,  1.])

    For greater speed, sending a pre-created numpy array as an argument
    avoids having to create a new one with each call:

    >>> grad = np.empty(grid.number_of_active_links)
    >>> rtn = grid.calc_grad_at_active_link(u, out=grad)
    >>> grad # doctest: +NORMALIZE_WHITESPACE
    array([ 1.,  1., -1.,
            1.,  1., -1.,  1.,
           -1., -1., -1.,
            1.,  1., -1.,  1.,
           -1.,  0.,  1.])
    >>> rtn is grad
    True

    >>> grid = RasterModelGrid((3, 3), spacing=(1, 2))
    >>> node_values = [0., 0., 0.,
    ...                1., 3., 1.,
    ...                2., 2., 2.]
    >>> grid.calc_grad_at_active_link(node_values)
    array([ 3.,  1., -1., -1.])

    This function is *deprecated*. Instead, use ``calc_grad_at_link``.

    >>> grid = RasterModelGrid((3, 3), spacing=(1, 2))
    >>> node_values = [0., 0., 0.,
    ...                1., 3., 1.,
    ...                2., 2., 2.]
    >>> grid.calc_grad_at_link(node_values)[grid.active_links]
    array([ 3.,  1., -1., -1.])

    LLCATS: LINF GRAD
    """
    if out is None:
        out = np.empty(len(grid.active_links), dtype=float)

    if len(out) != len(grid.active_links):
        raise ValueError('output buffer does not match that of the grid.')

    # grads = gradients.calculate_diff_at_active_links(grid, node_values,
    #                                                  out=out)
    grads = gradients.calc_diff_at_link(grid, node_values)
    out[:] = grads[grid.active_links]
    out /= grid.length_of_link[grid.active_links]

    return out


@use_field_name_or_array('node')
def calc_grad_across_cell_faces(grid, node_values, *args, **kwds):
    """calc_grad_across_cell_faces(grid, node_values, [cell_ids], out=None)

    Get gradients across the faces of a cell.

    Calculate gradient of the value field provided by *node_values* across
    each of the faces of the cells of a grid. The returned gradients are
    ordered as right, top, left, and bottom.

    Note that the returned gradients are masked to exclude neighbor nodes which
    are closed. Beneath the mask is the value -1.

    Parameters
    ----------
    grid : RasterModelGrid
        Source grid.
    node_values : array_like or field name
        Quantity to take the gradient of defined at each node.
    cell_ids : array_like, optional
        If provided, cell ids to measure gradients. Otherwise, find gradients
        for all cells.
    out : array_like, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    (N, 4) Masked ndarray
        Gradients for each face of the cell.

    Examples
    --------
    Create a grid with two cells.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 4)
    >>> x = np.array([0., 0., 0., 0.,
    ...               0., 0., 1., 1.,
    ...               3., 3., 3., 3.])

    A decrease in quantity across a face is a negative gradient.

    >>> grid.calc_grad_across_cell_faces(x) # doctest: +NORMALIZE_WHITESPACE
    masked_array(data =
     [[ 1.  3.  0.  0.]
     [ 0.  2. -1. -1.]],
                 mask =
     False,
           fill_value = 1e+20)

    >>> grid = RasterModelGrid((3, 4), spacing=(2, 1))
    >>> grid.calc_grad_across_cell_faces(x) # doctest: +NORMALIZE_WHITESPACE
    masked_array(data =
     [[ 1.   1.5  0.   0. ]
     [ 0.   1.  -1.  -0.5]],
                  mask =
     False,
           fill_value = 1e+20)

    LLCATS: FINF GRAD
    """
    padded_node_values = np.empty(node_values.size + 1, dtype=float)
    padded_node_values[-1] = BAD_INDEX_VALUE
    padded_node_values[:-1] = node_values
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]

    neighbors = grid.active_adjacent_nodes_at_node[node_ids]
    if BAD_INDEX_VALUE != -1:
        neighbors = np.where(neighbors == BAD_INDEX_VALUE, -1, neighbors)
    values_at_neighbors = padded_node_values[neighbors]
    masked_neighbor_values = np.ma.array(
        values_at_neighbors, mask=neighbors == BAD_INDEX_VALUE)
    values_at_nodes = node_values[node_ids].reshape(len(node_ids), 1)

    out = np.subtract(masked_neighbor_values, values_at_nodes, **kwds)

    out[:, (0, 2)] /= grid.dx
    out[:, (1, 3)] /= grid.dy

    return out


@use_field_name_or_array('node')
def calc_grad_across_cell_corners(grid, node_values, *args, **kwds):
    """calc_grad_across_cell_corners(grid, node_values, [cell_ids], out=None)

    Get gradients to diagonally opposite nodes.

    Calculate gradient of the value field provided by *node_values* to
    the values at diagonally opposite nodes. The returned gradients are
    ordered as upper-right, upper-left, lower-left and lower-right.

    Parameters
    ----------
    grid : RasterModelGrid
        Source grid.
    node_values : array_like or field name
        Quantity to take the gradient of defined at each node.
    cell_ids : array_like, optional
        If provided, cell ids to measure gradients. Otherwise, find gradients
        for all cells.
    out : array_like, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    (N, 4) ndarray
        Gradients to each diagonal node.

    Examples
    --------
    Create a grid with two cells.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 4)
    >>> x = np.array([1., 0., 0., 1.,
    ...               0., 0., 1., 1.,
    ...               3., 3., 3., 3.])

    A decrease in quantity to a diagonal node is a negative gradient.

    >>> from math import sqrt
    >>> grid.calc_grad_across_cell_corners(x) * sqrt(2.)
    array([[ 3.,  3.,  1.,  0.],
           [ 2.,  2., -1.,  0.]])

    >>> grid = RasterModelGrid((3, 4), spacing=(3, 4))
    >>> grid.calc_grad_across_cell_corners(x)
    array([[ 0.6,  0.6,  0.2,  0. ],
           [ 0.4,  0.4, -0.2,  0. ]])

    LLCATS: CNINF GRAD
    """
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]

    values_at_diagonals = node_values[grid.diagonal_adjacent_nodes_at_node[node_ids]]
    values_at_nodes = node_values[node_ids].reshape(len(node_ids), 1)

    out = np.subtract(values_at_diagonals, values_at_nodes, **kwds)
    np.divide(out, np.sqrt(grid.dy ** 2. + grid.dx ** 2.), out=out)

    return out


@use_field_name_or_array('node')
def calc_grad_along_node_links(grid, node_values, *args, **kwds):
    """calc_grad_along_node_links(grid, node_values, [cell_ids], out=None)
    
    Get gradients along links touching a node.

    Calculate gradient of the value field provided by *node_values* across
    each of the faces of the nodes of a grid. The returned gradients are
    ordered as right, top, left, and bottom. All returned values follow our
    standard sign convention, where a link pointing N or E and increasing in
    value is positive, a link pointing S or W and increasing in value is
    negative.

    Note that the returned gradients are masked to exclude neighbor nodes which
    are closed. Beneath the mask is the value -1.

    Parameters
    ----------
    grid : RasterModelGrid
        Source grid.
    node_values : array_like or field name
        Quantity to take the gradient of defined at each node.
    node_ids : array_like, optional
        If provided, node ids to measure gradients. Otherwise, find gradients
        for all nodes.
    out : array_like, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    (N, 4) Masked ndarray
        Gradients for each link of the node. Ordering is E,N,W,S.

    Examples
    --------
    Create a grid with nine nodes.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 3)
    >>> x = np.array([0., 0., 0.,
    ...               0., 1., 2.,
    ...               2., 2., 2.])

    A decrease in quantity across a face is a negative gradient.

    >>> grid.calc_grad_along_node_links(x) # doctest: +NORMALIZE_WHITESPACE
    masked_array(data =
     [[-- -- -- --]
     [-- 1.0 -- --]
     [-- -- -- --]
     [1.0 -- -- --]
     [1.0 1.0 1.0 1.0]
     [-- -- 1.0 --]
     [-- -- -- --]
     [-- -- -- 1.0]
     [-- -- -- --]],
                 mask =
     [[ True  True  True  True]
     [ True False  True  True]
     [ True  True  True  True]
     [False  True  True  True]
     [False False False False]
     [ True  True False  True]
     [ True  True  True  True]
     [ True  True  True False]
     [ True  True  True  True]],
           fill_value = 1e+20)

    >>> grid = RasterModelGrid((3, 3), spacing=(2, 4))
    >>> grid.calc_grad_along_node_links(x) # doctest: +NORMALIZE_WHITESPACE
    masked_array(data =
     [[-- -- -- --]
     [-- 0.5 -- --]
     [-- -- -- --]
     [0.25 -- -- --]
     [0.25 0.5 0.25 0.5]
     [-- -- 0.25 --]
     [-- -- -- --]
     [-- -- -- 0.5]
     [-- -- -- --]],
                 mask =
     [[ True  True  True  True]
     [ True False  True  True]
     [ True  True  True  True]
     [False  True  True  True]
     [False False False False]
     [ True  True False  True]
     [ True  True  True  True]
     [ True  True  True False]
     [ True  True  True  True]],
           fill_value = 1e+20)

    LLCATS: NINF LINF GRAD
    """
    padded_node_values = np.empty(node_values.size + 1, dtype=float)
    padded_node_values[-1] = BAD_INDEX_VALUE
    padded_node_values[:-1] = node_values
    node_ids = make_optional_arg_into_id_array(grid.number_of_nodes, *args)

    neighbors = grid.active_adjacent_nodes_at_node[node_ids]
    values_at_neighbors = padded_node_values[neighbors]
    masked_neighbor_values = np.ma.array(
        values_at_neighbors, mask=values_at_neighbors == BAD_INDEX_VALUE)
    values_at_nodes = node_values[node_ids].reshape(len(node_ids), 1)

    out = np.ma.empty_like(masked_neighbor_values, dtype=float)
    np.subtract(masked_neighbor_values[:, :2],
                values_at_nodes, out=out[:, :2], **kwds)
    np.subtract(values_at_nodes, masked_neighbor_values[:, 2:],
                out=out[:, 2:], **kwds)

    out[:, (0, 2)] /= grid.dx
    out[:, (1, 3)] /= grid.dy

    return out


def calc_unit_normals_at_cell_subtriangles(grid,
                                           elevs='topographic__elevation'):
    """Calculate unit normals on a cell.

    Calculate the eight unit normal vectors <a, b, c> to the eight
    subtriangles of a four-cornered (raster) cell.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.

    Returns
    -------
    (n_ENE, n_NNE, n_NNW, n_WNW, n_WSW, n_SSW, n_SSE, n_ESE) :
        each a num-cells x length-3 array
        Len-8 tuple of the eight unit normal vectors <a, b, c> for the eight
        subtriangles in the cell. Order is from north of east, counter
        clockwise to south of east (East North East, North North East, North
        North West, West North West, West South West, South South West, South
        South East, East South East).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((3, 3))
    >>> z = mg.node_x ** 2
    >>> eight_tris = mg.calc_unit_normals_at_cell_subtriangles(z)
    >>> type(eight_tris) is tuple
    True
    >>> len(eight_tris)
    8
    >>> eight_tris[0].shape == (mg.number_of_cells, 3)
    True
    >>> eight_tris # doctest: +NORMALIZE_WHITESPACE
    (array([[-0.9486833 ,  0.        ,  0.31622777]]),  
     array([[-0.9486833 ,  0.        ,  0.31622777]]), 
     array([[-0.70710678,  0.        ,  0.70710678]]), 
     array([[-0.70710678,  0.        ,  0.70710678]]), 
     array([[-0.70710678,  0.        ,  0.70710678]]), 
     array([[-0.70710678,  0.        ,  0.70710678]]), 
     array([[-0.9486833 ,  0.        ,  0.31622777]]), 
     array([[-0.9486833 ,  0.        ,  0.31622777]]))

    LLCATS: CINF GRAD
    """


    # identify the grid neigbors at each location
    I = grid.node_at_cell
    # calculate unit normals at all nodes.
    (n_ENE,
     n_NNE,
     n_NNW,
     n_WNW,
     n_WSW,
     n_SSW,
     n_SSE,
     n_ESE)=_calc_subtriangle_unit_normals_at_node(grid,elevs=elevs)

     # return only those at cell.
    return (n_ENE[I,:], 
            n_NNE[I,:], 
            n_NNW[I,:], 
            n_WNW[I,:],
            n_WSW[I,:],
            n_SSW[I,:],
            n_SSE[I,:],
            n_ESE[I,:])

def _calc_subtriangle_unit_normals_at_node(grid,
                                           elevs='topographic__elevation'):
    """Private Function: Calculate unit normals on subtriangles at all nodes.

    Calculate the eight unit normal vectors <a, b, c> to the eight
    subtriangles of a four-cornered (raster) cell. Unlike
    calc_unit_normals_at_node_subtriangles, this function also
    calculated unit normals at the degenerate part-cells around the
    boundary.
    
    On the grid boundaries where the cell is not fully defined, the unit normal
    is given as <nan, nan, nan>.

    Parameters
    ----------
    grid : RasterModelGrid
    A grid.
    elevs : str or ndarray, optional
    Field name or array of node values.

    Returns
    -------
    (n_ENE, n_NNE, n_NNW, n_WNW, n_WSW, n_SSW, n_SSE, n_ESE) :
    each a num-nodes x length-3 array
    Len-8 tuple of the eight unit normal vectors <a, b, c> for the eight
    subtriangles in the cell. Order is from north of east, counter
    clockwise to south of east (East North East, North North East, North
    North West, West North West, West South West, South South West, South
    South East, East South East).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.raster_gradients import(
    ...       _calc_subtriangle_unit_normals_at_node)
    >>> mg = RasterModelGrid((3, 3))
    >>> z = mg.node_x ** 2
    >>> eight_tris = _calc_subtriangle_unit_normals_at_node(mg, z)
    >>> type(eight_tris) is tuple
    True
    >>> len(eight_tris)
    8
    >>> eight_tris[0].shape == (mg.number_of_nodes, 3)
    True
    >>> eight_tris[0] # doctest: +NORMALIZE_WHITESPACE
    array([[-0.70710678,  0.        ,  0.70710678], 
               [-0.9486833 ,  0.        ,  0.31622777],
               [        nan,         nan,         nan],
               [-0.70710678,  0.        ,  0.70710678],
               [-0.9486833 ,  0.        ,  0.31622777],
               [        nan,         nan,         nan],
               [        nan,         nan,         nan],
               [        nan,         nan,         nan],
               [        nan,         nan,         nan]])

    LLCATS: CINF GRAD
    """
    
    try:
        z = grid.at_node[elevs]
    except TypeError:
        z = elevs
    #  cell has center node I
    # orthogonal neighbors P, R, T, V, counter clockwise from East
    # diagonal neihbors Q, S, U, W, counter clocwise from North East
    # There are 8 subtriangles that can be defined  with the following corners
    # (starting from the central node, and progressing counter-clockwise).
    # ENE: IPQ
    # NNE: IQR
    # NNW: IRS
    # WNW: IST
    # WSW: ITU
    # SSW: IUV
    # SSE: IVW
    # ESE: IWP

    # There are thus 8 vectors, IP, IQ, IR, IS, IT, IU, IV, IW

    # initialized difference matricies for cross product
    diff_xyz_IP = np.empty((grid.number_of_nodes, 3))  # East
    # ^this is the vector (xP-xI, yP-yI, zP-yI)
    diff_xyz_IQ = np.empty((grid.number_of_nodes, 3))  # Northeast
    diff_xyz_IR = np.empty((grid.number_of_nodes, 3))  # North
    diff_xyz_IS = np.empty((grid.number_of_nodes, 3))  # Northwest
    diff_xyz_IT = np.empty((grid.number_of_nodes, 3))  # West
    diff_xyz_IU = np.empty((grid.number_of_nodes, 3))  # Southwest
    diff_xyz_IV = np.empty((grid.number_of_nodes, 3))  # South
    diff_xyz_IW = np.empty((grid.number_of_nodes, 3))  # Southeast

    # identify the grid neigbors at each location
    I = np.arange(grid.number_of_nodes)
    P = grid.adjacent_nodes_at_node[I, 0]
    Q = grid.diagonal_adjacent_nodes_at_node[I, 0]
    R = grid.adjacent_nodes_at_node[I, 1]
    S = grid.diagonal_adjacent_nodes_at_node[I, 1]
    T = grid.adjacent_nodes_at_node[I, 2]
    U = grid.diagonal_adjacent_nodes_at_node[I, 2]
    V = grid.adjacent_nodes_at_node[I, 3]
    W = grid.diagonal_adjacent_nodes_at_node[I, 3]

    # get x, y, z coordinates for each location
    x_I = grid.node_x[I]
    y_I = grid.node_y[I]
    z_I = z[I]

    x_P = grid.node_x[P]
    y_P = grid.node_y[P]
    z_P = z[P]

    x_Q = grid.node_x[Q]
    y_Q = grid.node_y[Q]
    z_Q = z[Q]

    x_R = grid.node_x[R]
    y_R = grid.node_y[R]
    z_R = z[R]

    x_S = grid.node_x[S]
    y_S = grid.node_y[S]
    z_S = z[S]

    x_T = grid.node_x[T]
    y_T = grid.node_y[T]
    z_T = z[T]

    x_U = grid.node_x[U]
    y_U = grid.node_y[U]
    z_U = z[U]

    x_V = grid.node_x[V]
    y_V = grid.node_y[V]
    z_V = z[V]

    x_W = grid.node_x[W]
    y_W = grid.node_y[W]
    z_W = z[W]

    # calculate vectors by differencing
    diff_xyz_IP[:, 0] = x_P - x_I
    diff_xyz_IP[:, 1] = y_P - y_I
    diff_xyz_IP[:, 2] = z_P - z_I

    diff_xyz_IQ[:, 0] = x_Q - x_I
    diff_xyz_IQ[:, 1] = y_Q - y_I
    diff_xyz_IQ[:, 2] = z_Q - z_I

    diff_xyz_IR[:, 0] = x_R - x_I
    diff_xyz_IR[:, 1] = y_R - y_I
    diff_xyz_IR[:, 2] = z_R - z_I

    diff_xyz_IS[:, 0] = x_S - x_I
    diff_xyz_IS[:, 1] = y_S - y_I
    diff_xyz_IS[:, 2] = z_S - z_I

    diff_xyz_IT[:, 0] = x_T - x_I
    diff_xyz_IT[:, 1] = y_T - y_I
    diff_xyz_IT[:, 2] = z_T - z_I

    diff_xyz_IU[:, 0] = x_U - x_I
    diff_xyz_IU[:, 1] = y_U - y_I
    diff_xyz_IU[:, 2] = z_U - z_I

    diff_xyz_IV[:, 0] = x_V - x_I
    diff_xyz_IV[:, 1] = y_V - y_I
    diff_xyz_IV[:, 2] = z_V - z_I

    diff_xyz_IW[:, 0] = x_W - x_I
    diff_xyz_IW[:, 1] = y_W - y_I
    diff_xyz_IW[:, 2] = z_W - z_I

    # calculate cross product to get unit normal
    # cross product is orthogonal to both vectors, and is the normal
    # n = <a, b, c>, where plane is ax + by + cz = d
    nhat_ENE = np.cross(diff_xyz_IP, diff_xyz_IQ)  # <a, b, c>
    nhat_NNE = np.cross(diff_xyz_IQ, diff_xyz_IR)
    nhat_NNW = np.cross(diff_xyz_IR, diff_xyz_IS)
    nhat_WNW = np.cross(diff_xyz_IS, diff_xyz_IT)
    nhat_WSW = np.cross(diff_xyz_IT, diff_xyz_IU)
    nhat_SSW = np.cross(diff_xyz_IU, diff_xyz_IV)
    nhat_SSE = np.cross(diff_xyz_IV, diff_xyz_IW)
    nhat_ESE = np.cross(diff_xyz_IW, diff_xyz_IP)
    
     # now remove the bad subtriangles based on parts of the grid
    # make the bad subtriangle of length greater than one. 
    bad = np.nan*np.ones((3,))
    
    # first, corners:
    corners = grid.corner_nodes
    # lower left corner only has NNE and ENE
    nhat_NNW[corners[0],:] = bad
    nhat_WNW[corners[0],:] = bad
    nhat_WSW[corners[0],:] = bad
    nhat_SSW[corners[0],:] = bad
    nhat_SSE[corners[0],:] = bad
    nhat_ESE[corners[0],:] = bad
    
    # lower right corner only has NNW and WNW
    nhat_ENE[corners[1],:] = bad
    nhat_NNE[corners[1],:] = bad
    nhat_WSW[corners[1],:] = bad
    nhat_SSW[corners[1],:] = bad
    nhat_SSE[corners[1],:] = bad
    nhat_ESE[corners[1],:] = bad
    
    # upper left corner only has ESE and SSE
    nhat_ENE[corners[2],:] = bad
    nhat_NNE[corners[2],:] = bad
    nhat_NNW[corners[2],:] = bad
    nhat_WNW[corners[2],:] = bad
    nhat_WSW[corners[2],:] = bad
    nhat_SSW[corners[2],:] = bad

    
    # upper right corner only has WSW and SSW
    nhat_ENE[corners[3],:] = bad
    nhat_NNE[corners[3],:] = bad
    nhat_NNW[corners[3],:] = bad
    nhat_WNW[corners[3],:] = bad
    nhat_SSE[corners[3],:] = bad
    nhat_ESE[corners[3],:] = bad
    
    #next, sizes:
    # bottom row only has Norths
    bottom = grid.nodes_at_bottom_edge
    nhat_WSW[bottom, :] = bad
    nhat_SSW[bottom, :] = bad
    nhat_SSE[bottom, :] = bad
    nhat_ESE[bottom, :] = bad
    
    # left side only has Easts
    left = grid.nodes_at_left_edge
    nhat_NNW[left, :] = bad
    nhat_WNW[left, :] = bad
    nhat_WSW[left, :] = bad
    nhat_SSW[left, :] = bad
    
    # top row only has Souths
    top = grid.nodes_at_top_edge
    nhat_ENE[top, :] = bad
    nhat_NNE[top, :] = bad
    nhat_NNW[top, :] = bad
    nhat_WNW[top, :] = bad
    
    # right side only has Wests
    right = grid.nodes_at_right_edge
    nhat_ENE[right, :] = bad
    nhat_NNE[right, :] = bad
    nhat_SSE[right, :] = bad
    nhat_ESE[right, :] = bad
    
    # calculate magnitude of cross product so that the result is a unit normal
    nmag_ENE = np.sqrt(np.square(nhat_ENE).sum(axis=1))
    nmag_NNE = np.sqrt(np.square(nhat_NNE).sum(axis=1))
    nmag_NNW = np.sqrt(np.square(nhat_NNW).sum(axis=1))
    nmag_WNW = np.sqrt(np.square(nhat_WNW).sum(axis=1))
    nmag_WSW = np.sqrt(np.square(nhat_WSW).sum(axis=1))
    nmag_SSW = np.sqrt(np.square(nhat_SSW).sum(axis=1))
    nmag_SSE = np.sqrt(np.square(nhat_SSE).sum(axis=1))
    nmag_ESE = np.sqrt(np.square(nhat_ESE).sum(axis=1))

    # normalize the cross product with its magnitude so it is a unit normal
    # instead of a variable length normal.
    n_ENE = nhat_ENE/nmag_ENE.reshape(grid.number_of_nodes, 1)
    n_NNE = nhat_NNE/nmag_NNE.reshape(grid.number_of_nodes, 1)
    n_NNW = nhat_NNW/nmag_NNW.reshape(grid.number_of_nodes, 1)
    n_WNW = nhat_WNW/nmag_WNW.reshape(grid.number_of_nodes, 1)
    n_WSW = nhat_WSW/nmag_WSW.reshape(grid.number_of_nodes, 1)
    n_SSW = nhat_SSW/nmag_SSW.reshape(grid.number_of_nodes, 1)
    n_SSE = nhat_SSE/nmag_SSE.reshape(grid.number_of_nodes, 1)
    n_ESE = nhat_ESE/nmag_ESE.reshape(grid.number_of_nodes, 1)

    return (n_ENE, n_NNE, n_NNW, n_WNW, n_WSW, n_SSW, n_SSE, n_ESE)

def calc_slope_at_cell_subtriangles(grid, elevs='topographic__elevation',
                                    subtriangle_unit_normals=None):
    """Calculate the slope (positive magnitude of gradient) at each of the
    eight cell subtriangles.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    subtriangle_unit_normals : tuple of 8 (ncells, 3) arrays (optional)
        The unit normal vectors for the eight subtriangles of each cell,
        if already known. Order is from north of east, counter
        clockwise to south of east (East North East, North North East, North
        North West, West North West, West South West, South South West, South
        South East, East South East).

    Returns
    -------
    (s_ENE, s_NNE, s_NNW, s_WNW, s_WSW, s_SSW, s_SSE, s_ESE) :
        each a length num-cells array
        Len-8 tuple of the slopes (positive gradient magnitude) of each of the
        eight cell subtriangles, in radians. Order is from north of east,
        counter clockwise to south of east (East North East, North North East,
        North North West, West North West, West South West, South South West,
        South South East, East South East).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((3, 3))
    >>> z = np.array([np.sqrt(3.), 0., 4./3.,
    ...               0., 0., 0.,
    ...               1., 0., 1./np.sqrt(3.)])
    >>> eight_tris = mg.calc_unit_normals_at_cell_subtriangles(z)
    >>> S = mg.calc_slope_at_cell_subtriangles(z, eight_tris)
    >>> S0 = mg.calc_slope_at_cell_subtriangles(z)
    >>> np.allclose(S, S0)
    True
    >>> type(S) is tuple
    True
    >>> len(S)
    8
    >>> len(S[0]) == mg.number_of_cells
    True
    >>> np.allclose(S[0], S[1])
    True
    >>> np.allclose(S[2], S[3])
    True
    >>> np.allclose(S[4], S[5])
    True
    >>> np.allclose(S[6], S[7])
    True
    >>> np.allclose(np.rad2deg(S[0])[0], 30.)
    True
    >>> np.allclose(np.rad2deg(S[2])[0], 45.)
    True
    >>> np.allclose(np.rad2deg(S[4])[0], 60.)
    True
    >>> np.allclose(np.cos(S[6])[0], 3./5.)
    True

    LLCATS: CINF GRAD
    """

    # calculate all subtriangle slopes
    (s_ENE,
     s_NNE,
     s_NNW,
     s_WNW,
     s_WSW,
     s_SSW,
     s_SSE,
     s_ESE) = _calc_subtriangle_slopes_at_node(grid, elevs=elevs, 
                                               subtriangle_unit_normals=subtriangle_unit_normals)
    # return only those at cell
    if s_ENE.shape[0] == grid.number_of_nodes:
        I = grid.node_at_cell
    else:
        I = np.arange(grid.number_of_cells)
        
    return (s_ENE[I], s_NNE[I], s_NNW[I], s_WNW[I], s_WSW[I], s_SSW[I], s_SSE[I], s_ESE[I])

def _calc_subtriangle_slopes_at_node(grid, elevs='topographic__elevation',
                                     subtriangle_unit_normals=None):
    """Private Function: Calculate subtriangles slope at all nodes. 
    
    Calculate the slope (positive magnitude of gradient) at each of the
    eight subtriangles, including those at not-full cells along the
    boundary. 
    
    Those subtriangles that that don't exist because they are on the edge
    of the grid have slopes of NAN.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    subtriangle_unit_normals : tuple of 8 (ncells, 3) or (nnodes, 3) arrays 
        (optional)
        The unit normal vectors for the eight subtriangles of each cell or 
        node,if already known. Order is from north of east, counter
        clockwise to south of east (East North East, North North East, North
        North West, West North West, West South West, South South West, South
        South East, East South East).

    Returns
    -------
    (s_ENE, s_NNE, s_NNW, s_WNW, s_WSW, s_SSW, s_SSE, s_ESE) :
        each a length num-cells array
        Len-8 tuple of the slopes (positive gradient magnitude) of each of the
        eight cell subtriangles, in radians. Order is from north of east,
        counter clockwise to south of east (East North East, North North East,
        North North West, West North West, West South West, South South West,
        South South East, East South East).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.raster_gradients import(
    ...                         _calc_subtriangle_unit_normals_at_node,
    ...                         _calc_subtriangle_slopes_at_node)
    >>> mg = RasterModelGrid((3, 3))
    >>> z = np.array([np.sqrt(3.), 0., 4./3.,
    ...               0., 0., 0.,
    ...               1., 0., 1./np.sqrt(3.)])
    >>> eight_tris = _calc_subtriangle_unit_normals_at_node(mg, z)
    >>> S = _calc_subtriangle_slopes_at_node(mg, z, eight_tris)
    >>> S0 = _calc_subtriangle_slopes_at_node(mg, z)
    >>> np.allclose(S, S0, equal_nan=True)
    True
    >>> type(S) is tuple
    True
    >>> len(S)
    8
    >>> len(S[0]) == mg.number_of_nodes
    True
    >>> np.allclose(S[0][mg.core_nodes], S[1][mg.core_nodes])
    True
    >>> np.allclose(S[2][mg.core_nodes], S[3][mg.core_nodes])
    True
    >>> np.allclose(S[4][mg.core_nodes], S[5][mg.core_nodes])
    True
    >>> np.allclose(S[6][mg.core_nodes], S[7][mg.core_nodes])
    True
    >>> np.allclose(np.rad2deg(S[0][mg.core_nodes]), 30.)
    True
    >>> np.allclose(np.rad2deg(S[2][mg.core_nodes]), 45.)
    True
    >>> np.allclose(np.rad2deg(S[4])[mg.core_nodes], 60.)
    True
    >>> np.allclose(np.cos(S[6])[mg.core_nodes], 3./5.)
    True

    LLCATS: CINF GRAD
    """

    # verify that subtriangle_unit_normals is of the correct form.
    if subtriangle_unit_normals is not None:
        assert len(subtriangle_unit_normals) == 8
        assert subtriangle_unit_normals[0].shape[1] == 3
        assert subtriangle_unit_normals[1].shape[1] == 3
        assert subtriangle_unit_normals[2].shape[1] == 3
        assert subtriangle_unit_normals[3].shape[1] == 3
        assert subtriangle_unit_normals[4].shape[1] == 3
        assert subtriangle_unit_normals[5].shape[1] == 3
        assert subtriangle_unit_normals[6].shape[1] == 3
        assert subtriangle_unit_normals[7].shape[1] == 3
        (n_ENE, n_NNE, n_NNW, n_WNW,
        n_WSW, n_SSW, n_SSE, n_ESE) = subtriangle_unit_normals
        
        if subtriangle_unit_normals[7].shape[0] == grid.number_of_nodes:
            reshape_size = grid.number_of_nodes
        elif subtriangle_unit_normals[7].shape[0] == grid.number_of_cells:
            reshape_size = grid.number_of_cells              
        else:
            ValueError("Subtriangles must be of lenght nnodes or ncells")    
    else:
        n_ENE, n_NNE, n_NNW, n_WNW, n_WSW, n_SSW, n_SSE, n_ESE = (
            _calc_subtriangle_unit_normals_at_node(grid, elevs))
        reshape_size = grid.number_of_nodes
        
    # combine z direction element of all eight so that the arccosine portion
    # only takes one function call.
    dotprod = np.empty((reshape_size, 8))
    dotprod[:, 0] = n_ENE[:, 2]  # by definition
    dotprod[:, 1] = n_NNE[:, 2]
    dotprod[:, 2] = n_NNW[:, 2]
    dotprod[:, 3] = n_WNW[:, 2]
    dotprod[:, 4] = n_WSW[:, 2]
    dotprod[:, 5] = n_SSW[:, 2]
    dotprod[:, 6] = n_SSE[:, 2]
    dotprod[:, 7] = n_ESE[:, 2]

    # take the inverse cosine of the z component to get the slope angle
    slopes_at_cell_subtriangles = np.arccos(dotprod)  #

    # split array into each subtriangle component.
    s_ENE = slopes_at_cell_subtriangles[:, 0].reshape(reshape_size)
    s_NNE = slopes_at_cell_subtriangles[:, 1].reshape(reshape_size)
    s_NNW = slopes_at_cell_subtriangles[:, 2].reshape(reshape_size)
    s_WNW = slopes_at_cell_subtriangles[:, 3].reshape(reshape_size)
    s_WSW = slopes_at_cell_subtriangles[:, 4].reshape(reshape_size)
    s_SSW = slopes_at_cell_subtriangles[:, 5].reshape(reshape_size)
    s_SSE = slopes_at_cell_subtriangles[:, 6].reshape(reshape_size)
    s_ESE = slopes_at_cell_subtriangles[:, 7].reshape(reshape_size)

    return (s_ENE, s_NNE, s_NNW, s_WNW, s_WSW, s_SSW, s_SSE, s_ESE)

def calc_aspect_at_cell_subtriangles(grid, elevs='topographic__elevation',
                                     subtriangle_unit_normals=None,
                                     unit='degrees'):
    """Get tuple of arrays of aspect of each of the eight cell subtriangles.

    Aspect is returned as radians clockwise of north, unless input parameter
    units is set to 'degrees'.

    If subtriangle_unit_normals is provided the aspect will be calculated from
    these data.

    If it is not, it will be derived from elevation data at the nodes,
    which can either be a string referring to a grid field (default:
    'topographic__elevation'), or an nnodes-long numpy array of the
    values themselves.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    elevs : str or array (optional)
        Node field name or node array of elevations.
        If *subtriangle_unit_normals* is not provided, must be set, but unused
        otherwise.
    subtriangle_unit_normals : tuple of 8 (ncels, 3) arrays (optional)
        The unit normal vectors for the eight subtriangles of each cell,
        if already known. Order is from north of east, counter
        clockwise to south of east (East North East, North North East, North
        North West, West North West, West South West, South South West, South
        South East, East South East).
    unit : {'degrees', 'radians'}
        Controls the unit that the aspect is returned as.

    Returns
    -------
    (a_ENE, a_NNE, a_NNW, a_WNW, a_WSW, a_SSW, a_SSE, a_ESE) :
            each a length num-cells array
        Len-8 tuple of the aspect of each of the eight cell subtriangles.
        Aspect is returned as angle clockwise of north. Units are given as
        radians unless input parameter units is set to 'degrees'.
        Order is from north of east, counter clockwise to south of east (East
        North East, North North East, North North West, West North West, West
        South West, South South West, South South East, East South East).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((3, 3))
    >>> z = np.array([1., 0., 1., 0., 0., 0., 1., 0., 1.])
    >>> eight_tris = mg.calc_unit_normals_at_cell_subtriangles(z)
    >>> A = mg.calc_aspect_at_cell_subtriangles(z, eight_tris)
    >>> A0 = mg.calc_aspect_at_cell_subtriangles(z)
    >>> np.allclose(A, A0)
    True
    >>> type(A) is tuple
    True
    >>> len(A)
    8
    >>> len(A[0]) == mg.number_of_cells
    True
    >>> A0  # doctest: +NORMALIZE_WHITESPACE
    (array([ 180.]), array([ 270.]), array([ 90.]), array([ 180.]),
     array([ 0.]), array([ 90.]), array([ 270.]), array([ 0.]))


    LLCATS: CINF SURF
    """

     # calculate all subtriangle slopes
    (angle_ENE,
     angle_NNE,
     angle_NNW,
     angle_WNW,
     angle_WSW,
     angle_SSW,
     angle_SSE,
     angle_ESE) = _calc_subtriangle_aspect_at_node(grid, elevs=elevs, 
                                                   subtriangle_unit_normals=subtriangle_unit_normals,
                                                   unit=unit)
    # return only those at cell
    if angle_ESE.shape[0] == grid.number_of_nodes:
        I = grid.node_at_cell
    else:
        I = np.arange(grid.number_of_cells)

    if unit == 'degrees' or unit == 'radians':
        return (angle_ENE[I],
                angle_NNE[I],
                angle_NNW[I],
                angle_WNW[I],
                angle_WSW[I],
                angle_SSW[I],
                angle_SSE[I],
                angle_ESE[I])
    else:
        raise TypeError("unit must be 'degrees' or 'radians'")


def _calc_subtriangle_aspect_at_node(grid, elevs='topographic__elevation',
                                     subtriangle_unit_normals=None,
                                     unit='degrees'):
    """Private Function: Aspect of subtriangles at node. 
    
    This function calculates the aspect of all subtriangles, including those
    that are at noded without cells (on the boundaries).
    
    Aspect is returned as radians clockwise of north, unless input parameter
    units is set to 'degrees'.

    If subtriangle_unit_normals is provided the aspect will be calculated from
    these data.

    If it is not, it will be derived from elevation data at the nodes,
    which can either be a string referring to a grid field (default:
    'topographic__elevation'), or an nnodes-long numpy array of the
    values themselves.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    elevs : str or array (optional)
        Node field name or node array of elevations.
        If *subtriangle_unit_normals* is not provided, must be set, but unused
        otherwise.
    subtriangle_unit_normals : tuple of 8 (ncells, 3) or (nnodes, 3) arrays 
        (optional)
        The unit normal vectors for the eight subtriangles of each cell or 
        node,if already known. Order is from north of east, counter
        clockwise to south of east (East North East, North North East, North
        North West, West North West, West South West, South South West, South
        South East, East South East).
    unit : {'degrees', 'radians'}
        Controls the unit that the aspect is returned as.

    Returns
    -------
    (a_ENE, a_NNE, a_NNW, a_WNW, a_WSW, a_SSW, a_SSE, a_ESE) :
            each a length num-cells array
        Len-8 tuple of the aspect of each of the eight cell subtriangles.
        Aspect is returned as angle clockwise of north. Units are given as
        radians unless input parameter units is set to 'degrees'.
        Order is from north of east, counter clockwise to south of east (East
        North East, North North East, North North West, West North West, West
        South West, South South West, South South East, East South East).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.raster_gradients import (
    ...     _calc_subtriangle_unit_normals_at_node,
    ...     _calc_subtriangle_aspect_at_node)
    >>> mg = RasterModelGrid((3, 3))
    >>> z = np.array([1., 0., 1., 0., 0., 0., 1., 0., 1.])
    >>> eight_tris = _calc_subtriangle_unit_normals_at_node(mg, z)
    >>> A = _calc_subtriangle_aspect_at_node(mg, z, eight_tris)
    >>> A0 = _calc_subtriangle_aspect_at_node(mg, z)
    >>> np.allclose(A, A0, equal_nan=True)
    True
    >>> type(A) is tuple
    True
    >>> len(A)
    8
    >>> len(A[0]) == mg.number_of_nodes
    True
    >>> A0  # doctest: +NORMALIZE_WHITESPACE
    (array([  90.,  315.,   nan,   90.,  180.,   nan,   nan,   nan,   nan]), 
     array([   0.,   90.,   nan,  135.,  270.,   nan,   nan,   nan,   nan]), 
     array([  nan,   90.,    0.,   nan,   90.,  225.,   nan,   nan,   nan]), 
     array([  nan,   45.,  270.,   nan,  180.,   90.,   nan,   nan,   nan]), 
     array([  nan,   nan,   nan,   nan,    0.,   90.,   nan,  135.,  270.]), 
     array([  nan,   nan,   nan,   nan,   90.,  315.,   nan,   90.,  180.]), 
     array([  nan,   nan,   nan,   45.,  270.,   nan,  180.,   90.,   nan]), 
     array([  nan,   nan,   nan,  270.,    0.,   nan,   90.,  225.,   nan]))

    LLCATS: CINF SURF
    """

    # verify that subtriangle_unit_normals is of the correct form.
    if subtriangle_unit_normals is not None:
        assert len(subtriangle_unit_normals) == 8
        assert subtriangle_unit_normals[0].shape[1] == 3
        assert subtriangle_unit_normals[1].shape[1] == 3
        assert subtriangle_unit_normals[2].shape[1] == 3
        assert subtriangle_unit_normals[3].shape[1] == 3
        assert subtriangle_unit_normals[4].shape[1] == 3
        assert subtriangle_unit_normals[5].shape[1] == 3
        assert subtriangle_unit_normals[6].shape[1] == 3
        assert subtriangle_unit_normals[7].shape[1] == 3
                                       
        if subtriangle_unit_normals[7].shape[0] == grid.number_of_nodes:
            reshape_size = grid.number_of_nodes
        elif subtriangle_unit_normals[7].shape[0] == grid.number_of_cells:
            reshape_size = grid.number_of_cells              
        else:
            ValueError("Subtriangles must be of lenght nnodes or ncells")
        (n_ENE, n_NNE, n_NNW, n_WNW,
         n_WSW, n_SSW, n_SSE, n_ESE) = subtriangle_unit_normals

    # otherwise create it.
    else:
        n_ENE, n_NNE, n_NNW, n_WNW, n_WSW, n_SSW, n_SSE, n_ESE = (
            _calc_subtriangle_unit_normals_at_node(grid, elevs))
        reshape_size = grid.number_of_nodes
    # calculate the aspect as an angle ccw from the x axis (math angle)
    angle_from_x_ccw_ENE = np.reshape(np.arctan2(n_ENE[:, 1], n_ENE[:, 0]),
                                      reshape_size)
    angle_from_x_ccw_NNE = np.reshape(np.arctan2(n_NNE[:, 1], n_NNE[:, 0]),
                                      reshape_size)
    angle_from_x_ccw_NNW = np.reshape(np.arctan2(n_NNW[:, 1], n_NNW[:, 0]),
                                      reshape_size)
    angle_from_x_ccw_WNW = np.reshape(np.arctan2(n_WNW[:, 1], n_WNW[:, 0]),
                                      reshape_size)
    angle_from_x_ccw_WSW = np.reshape(np.arctan2(n_WSW[:, 1], n_WSW[:, 0]),
                                      reshape_size)
    angle_from_x_ccw_SSW = np.reshape(np.arctan2(n_SSW[:, 1], n_SSW[:, 0]),
                                      reshape_size)
    angle_from_x_ccw_SSE = np.reshape(np.arctan2(n_SSE[:, 1], n_SSE[:, 0]),
                                      reshape_size)
    angle_from_x_ccw_ESE = np.reshape(np.arctan2(n_ESE[:, 1], n_ESE[:, 0]),
                                      reshape_size)
    # convert reference from math angle to angles clockwise from north
    # return as either  radians or degrees depending on unit.
    if unit == 'degrees':
        return (radians_to_degrees(angle_from_x_ccw_ENE),
                radians_to_degrees(angle_from_x_ccw_NNE),
                radians_to_degrees(angle_from_x_ccw_NNW),
                radians_to_degrees(angle_from_x_ccw_WNW),
                radians_to_degrees(angle_from_x_ccw_WSW),
                radians_to_degrees(angle_from_x_ccw_SSW),
                radians_to_degrees(angle_from_x_ccw_SSE),
                radians_to_degrees(angle_from_x_ccw_ESE))

    elif unit == 'radians':
        angle_from_north_cw_ENE = (5. * np.pi / 2. -
                                   angle_from_x_ccw_ENE) % (2. * np.pi)
        angle_from_north_cw_NNE = (5. * np.pi / 2. -
                                   angle_from_x_ccw_NNE) % (2. * np.pi)
        angle_from_north_cw_NNW = (5. * np.pi / 2. -
                                   angle_from_x_ccw_NNW) % (2. * np.pi)
        angle_from_north_cw_WNW = (5. * np.pi / 2. -
                                   angle_from_x_ccw_WNW) % (2. * np.pi)
        angle_from_north_cw_WSW = (5. * np.pi / 2. -
                                   angle_from_x_ccw_WSW) % (2. * np.pi)
        angle_from_north_cw_SSW = (5. * np.pi / 2. -
                                   angle_from_x_ccw_SSW) % (2. * np.pi)
        angle_from_north_cw_SSE = (5. * np.pi / 2. -
                                   angle_from_x_ccw_SSE) % (2. * np.pi)
        angle_from_north_cw_ESE = (5. * np.pi / 2. -
                                   angle_from_x_ccw_ESE) % (2. * np.pi)

        return (angle_from_north_cw_ENE,
                angle_from_north_cw_NNE,
                angle_from_north_cw_NNW,
                angle_from_north_cw_WNW,
                angle_from_north_cw_WSW,
                angle_from_north_cw_SSW,
                angle_from_north_cw_SSE,
                angle_from_north_cw_ESE)
    else:
        raise TypeError("unit must be 'degrees' or 'radians'")
        
def calc_unit_normals_at_patch_subtriangles(grid,
                                            elevs='topographic__elevation'):
    """Calculate unit normals on a patch.

    Calculate the four unit normal vectors <a, b, c> to the four possible
    subtriangles of a four-cornered (raster) patch.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.

    Returns
    -------
    (n_TR, n_TL, n_BL, n_BR) : each a num-patches x length-3 array
        Len-4 tuple of the four unit normal vectors <a, b, c> for the four
        possible subtriangles in the patch. Order is (topright, topleft,
        bottomleft, bottomright).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.node_x ** 2
    >>> four_tris = mg.calc_unit_normals_at_patch_subtriangles(z)
    >>> type(four_tris) is tuple
    True
    >>> len(four_tris)
    4
    >>> np.allclose(four_tris[0], four_tris[1])
    True
    >>> np.allclose(four_tris[2], four_tris[3])
    True
    >>> np.allclose(four_tris[0], four_tris[2])
    True
    >>> np.allclose(np.square(four_tris[0]).sum(axis=1), 1.)
    True
    >>> four_tris[0]
    array([[-0.70710678,  0.        ,  0.70710678],
           [-0.9486833 ,  0.        ,  0.31622777],
           [-0.98058068,  0.        ,  0.19611614],
           [-0.98994949,  0.        ,  0.14142136],
           [-0.70710678,  0.        ,  0.70710678],
           [-0.9486833 ,  0.        ,  0.31622777],
           [-0.98058068,  0.        ,  0.19611614],
           [-0.98994949,  0.        ,  0.14142136],
           [-0.70710678,  0.        ,  0.70710678],
           [-0.9486833 ,  0.        ,  0.31622777],
           [-0.98058068,  0.        ,  0.19611614],
           [-0.98994949,  0.        ,  0.14142136]])

    LLCATS: PINF GRAD
    """
    try:
        z = grid.at_node[elevs]
    except TypeError:
        z = elevs
    # conceptualize patches as TWO sets of 3 nodes
    # the corners are PQRS, CC from NE
    diff_xyz_PQ = np.empty((grid.number_of_patches, 3))  # TOP
    # ^this is the vector (xQ-xP, yQ-yP, zQ-yP)
    diff_xyz_PS = np.empty((grid.number_of_patches, 3))  # RIGHT
    # we have RS and QR implicitly in PQ and PS - but store them too
    diff_xyz_RS = np.empty((grid.number_of_patches, 3))  # BOTTOM
    diff_xyz_QR = np.empty((grid.number_of_patches, 3))  # LEFT
    P = grid.nodes_at_patch[:, 0]
    Q = grid.nodes_at_patch[:, 1]
    R = grid.nodes_at_patch[:, 2]
    S = grid.nodes_at_patch[:, 3]
    x_P = grid.node_x[P]
    y_P = grid.node_y[P]
    z_P = z[P]
    x_Q = grid.node_x[Q]
    y_Q = grid.node_y[Q]
    z_Q = z[Q]
    x_R = grid.node_x[R]
    y_R = grid.node_y[R]
    z_R = z[R]
    x_S = grid.node_x[S]
    y_S = grid.node_y[S]
    z_S = z[S]
    diff_xyz_PQ[:, 0] = x_Q - x_P
    diff_xyz_PQ[:, 1] = y_Q - y_P
    diff_xyz_PQ[:, 2] = z_Q - z_P
    diff_xyz_PS[:, 0] = x_S - x_P
    diff_xyz_PS[:, 1] = y_S - y_P
    diff_xyz_PS[:, 2] = z_S - z_P
    diff_xyz_RS[:, 0] = x_S - x_R
    diff_xyz_RS[:, 1] = y_S - y_R
    diff_xyz_RS[:, 2] = z_S - z_R
    diff_xyz_QR[:, 0] = x_R - x_Q
    diff_xyz_QR[:, 1] = y_R - y_Q
    diff_xyz_QR[:, 2] = z_R - z_Q
    # make the other ones
    # cross product is orthogonal to both vectors, and is the normal
    # n = <a, b, c>, where plane is ax + by + cz = d
    nhat_topleft = np.cross(diff_xyz_PQ, diff_xyz_QR)  # <a, b, c>
    nhat_bottomright = np.cross(diff_xyz_PS, diff_xyz_RS)
    nhat_topright = np.cross(diff_xyz_PQ, diff_xyz_PS)
    nhat_bottomleft = np.cross(diff_xyz_QR, diff_xyz_RS)
    nmag_topleft = np.sqrt(np.square(nhat_topleft).sum(axis=1))
    nmag_bottomright = np.sqrt(np.square(nhat_bottomright).sum(axis=1))
    nmag_topright = np.sqrt(np.square(nhat_topright).sum(axis=1))
    nmag_bottomleft = np.sqrt(np.square(nhat_bottomleft).sum(axis=1))
    n_TR = nhat_topright/nmag_topright.reshape(grid.number_of_patches, 1)
    n_TL = nhat_topleft/nmag_topleft.reshape(grid.number_of_patches, 1)
    n_BL = nhat_bottomleft/nmag_bottomleft.reshape(
        grid.number_of_patches, 1)
    n_BR = nhat_bottomright/nmag_bottomright.reshape(
        grid.number_of_patches, 1)

    return (n_TR, n_TL, n_BL, n_BR)


def calc_slope_at_patch(grid, elevs='topographic__elevation',
                        ignore_closed_nodes=True,
                        subtriangle_unit_normals=None):
    """Calculate the slope (positive magnitude of gradient) at raster patches.

    Returns the mean of the slopes of the four possible patch subtriangles.

    If ignore_closed_nodes is True, closed nodes do not affect slope
    calculations. If more than one closed node is present in a patch, the
    patch slope is set to zero.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    ignore_closed_nodes : bool
        If True, do not incorporate values at closed nodes into the calc.
    subtriangle_unit_normals : tuple of 4 (npatches, 3) arrays (optional)
        The unit normal vectors for the four subtriangles of each patch,
        if already known. Order is TR, TL, BL, BR.

    Returns
    -------
    slopes_at_patch : n_patches-long array
        The slope (positive gradient magnitude) of each patch, in radians.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.node_x
    >>> S = mg.calc_slope_at_patch(elevs=z)
    >>> S.size == mg.number_of_patches
    True
    >>> np.allclose(S, np.pi/4.)
    True
    >>> z = mg.node_y**2
    >>> mg.calc_slope_at_patch(elevs=z).reshape((3, 4))
    array([[ 0.78539816,  0.78539816,  0.78539816,  0.78539816],
           [ 1.24904577,  1.24904577,  1.24904577,  1.24904577],
           [ 1.37340077,  1.37340077,  1.37340077,  1.37340077]])

    >>> from landlab import CLOSED_BOUNDARY, FIXED_VALUE_BOUNDARY
    >>> z = mg.node_x.copy()
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> mg.status_at_node[11] = CLOSED_BOUNDARY
    >>> mg.status_at_node[9] = FIXED_VALUE_BOUNDARY
    >>> z[11] = 100.  # this should get ignored now
    >>> z[9] = 2.  # this should be felt by patch 7 only
    >>> mg.calc_slope_at_patch(elevs=z, ignore_closed_nodes=True).reshape(
    ...     (3, 4)) * 4./np.pi
    array([[ 0.,  0.,  0.,  0.],
           [ 0.,  1.,  1.,  1.],
           [ 0.,  0.,  0.,  0.]])

    LLCATS: PINF GRAD
    """
    if subtriangle_unit_normals is not None:
        assert len(subtriangle_unit_normals) == 4
        assert subtriangle_unit_normals[0].shape[1] == 3
        assert subtriangle_unit_normals[1].shape[1] == 3
        assert subtriangle_unit_normals[2].shape[1] == 3
        assert subtriangle_unit_normals[3].shape[1] == 3
        n_TR, n_TL, n_BL, n_BR = subtriangle_unit_normals
    else:
        n_TR, n_TL, n_BL, n_BR = (
            grid.calc_unit_normals_at_patch_subtriangles(elevs))
    dotprod_TL = n_TL[:, 2]  # by definition
    dotprod_BR = n_BR[:, 2]
    dotprod_TR = n_TR[:, 2]
    dotprod_BL = n_BL[:, 2]
    slopes_at_patch_TL = np.arccos(dotprod_TL)  # 1 node order
    slopes_at_patch_BR = np.arccos(dotprod_BR)  # 3
    slopes_at_patch_TR = np.arccos(dotprod_TR)  # 0
    slopes_at_patch_BL = np.arccos(dotprod_BL)  # 2
    if ignore_closed_nodes:
        badnodes = grid.status_at_node[grid.nodes_at_patch] == CLOSED_BOUNDARY
        tot_bad = badnodes.sum(axis=1)
        tot_tris = 4. - 3. * (tot_bad > 0)  # 4 where all good, 1 where not
        # now shut down the bad tris. Remember, one bad node => 3 bad tris.
        # anywhere where badnodes > 1 will have zero from summing, so div by 1
        # assert np.all(np.logical_or(np.isclose(tot_tris, 4.),
        #                             np.isclose(tot_tris, 1.)))
        corners_rot = deque([slopes_at_patch_BR, slopes_at_patch_TR,
                             slopes_at_patch_TL, slopes_at_patch_BL])
        # note initial offset so we are centered around TR on first slice
        for i in range(4):
            for j in range(3):
                (corners_rot[j])[badnodes[:, i]] = 0.
            corners_rot.rotate(-1)
    else:
        tot_tris = 4.
    mean_slope_at_patch = (slopes_at_patch_TR + slopes_at_patch_TL +
                           slopes_at_patch_BL + slopes_at_patch_BR) / tot_tris

    return mean_slope_at_patch


def calc_grad_at_patch(grid, elevs='topographic__elevation',
                       ignore_closed_nodes=True,
                       subtriangle_unit_normals=None,
                       slope_magnitude=None):
    """Calculate the components of the gradient of each raster patch.

    Returns the mean gradient of the four possible patch subtriangles,
    in radians.

    If ignore_closed_nodes is True, closed nodes do not affect gradient
    calculations. If more than one closed node is present in a patch, the
    patch gradients in both x and y directions are set to zero.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    ignore_closed_nodes : bool
        If True, do not incorporate values at closed nodes into the calc.
    subtriangle_unit_normals : tuple of 4 (npatches, 3) arrays (optional)
        The unit normal vectors for the four subtriangles of each patch,
        if already known. Order is TR, TL, BL, BR.
    slope_magnitude : array with size num_patches (optional)
        The mean slope of each patch, if already known. Units must be the
        same as provided here!

    Returns
    -------
    gradient_tuple : (x_component_at_patch, y_component_at_patch)
        Len-2 tuple of arrays giving components of gradient in the x and y
        directions, in the units of *radians*.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.node_y
    >>> (x_grad, y_grad) = mg.calc_grad_at_patch(elevs=z)
    >>> np.allclose(y_grad, np.pi/4.)
    True
    >>> np.allclose(x_grad, 0.)
    True

    >>> from landlab import CLOSED_BOUNDARY, FIXED_VALUE_BOUNDARY
    >>> z = mg.node_x.copy()
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> mg.status_at_node[11] = CLOSED_BOUNDARY
    >>> mg.status_at_node[[9, 2]] = FIXED_VALUE_BOUNDARY
    >>> z[11] = 100.  # this should get ignored now
    >>> z[9] = 2.  # this should be felt by patch 7 only
    >>> z[2] = 1.  # should be felt by patches 1 and 2
    >>> xgrad, ygrad = mg.calc_grad_at_patch(
    ...     elevs=z, ignore_closed_nodes=True)
    >>> (xgrad.reshape((3, 4)) * 4./np.pi)[1, 1:]
    array([ 1.,  1., -1.])
    >>> np.allclose(ygrad[1:3], xgrad[1:3])
    True

    LLCATS: PINF GRAD
    """
    if subtriangle_unit_normals is not None:
        assert len(subtriangle_unit_normals) == 4
        assert subtriangle_unit_normals[0].shape[1] == 3
        assert subtriangle_unit_normals[1].shape[1] == 3
        assert subtriangle_unit_normals[2].shape[1] == 3
        assert subtriangle_unit_normals[3].shape[1] == 3
        n_TR, n_TL, n_BL, n_BR = subtriangle_unit_normals
    else:
        n_TR, n_TL, n_BL, n_BR = \
            grid.calc_unit_normals_at_patch_subtriangles(elevs)
    if slope_magnitude is not None:
        assert slope_magnitude.size == grid.number_of_patches
        slopes_at_patch = slope_magnitude
    else:
        slopes_at_patch = grid.calc_slope_at_patch(
            elevs=elevs, ignore_closed_nodes=ignore_closed_nodes,
            subtriangle_unit_normals=(n_TR, n_TL, n_BL, n_BR))

    if ignore_closed_nodes:
        badnodes = grid.status_at_node[grid.nodes_at_patch] == CLOSED_BOUNDARY
        corners_rot = deque([n_BR, n_TR, n_TL, n_BL])
        # note initial offset so we are centered around TR on first slice
        for i in range(4):
            for j in range(3):
                (corners_rot[j])[badnodes[:, i], :] = 0.
            corners_rot.rotate(-1)

    n_sum_x = n_TR[:, 0] + n_TL[:, 0] + n_BL[:, 0] + n_BR[:, 0]
    n_sum_y = n_TR[:, 1] + n_TL[:, 1] + n_BL[:, 1] + n_BR[:, 1]
    theta_sum = np.arctan2(-n_sum_y, -n_sum_x)
    x_slope_patches = np.cos(theta_sum)*slopes_at_patch
    y_slope_patches = np.sin(theta_sum)*slopes_at_patch

    return (x_slope_patches, y_slope_patches)


def calc_slope_at_node(grid, elevs='topographic__elevation',
                       method='patch_mean', ignore_closed_nodes=True,
                       return_components=False):
    """Array of slopes at nodes, averaged over neighboring patches.

    Produces a value for node slope (i.e., mean gradient magnitude)
    at each node in a manner analogous to a GIS-style slope map.
    If method=='patch_mean', it averages the gradient on each of the
    patches surrounding the node; if method=='Horn', it returns the
    resolved slope direction. Directional information can still be
    returned through use of the return_components keyword.
    All values are returned in radians, including the components;
    take the tan to recover the rise/run.

    Note that under these definitions, it is not always true that::

        mag, cmp = mg.calc_slope_at_node(z)
        mag**2 == cmp[0]**2 + cmp[1]**2  # only if method=='Horn'

    If ignore_closed_nodes is False, all proximal elevation values will be used
    in the calculation. If True, only unclosed nodes are used.

    This is a verion of this code specialized for a raster. It subdivides
    the four square patches around each node into subtriangles,
    in order to ensure more correct solutions that incorporate equally
    weighted information from all surrounding nodes on rough surfaces.

    Parameters
    ----------
    elevs : str or ndarray, optional
        Field name or array of node values.
    method : {'patch_mean', 'Horn'}
        Controls the slope algorithm. Current options are 'patch_mean',
        which takes the mean slope of each pf the four neighboring
        square patches, and 'Horn', which is the standard ArcGIS slope
        algorithm. These produce very similar solutions; the Horn method
        gives a vector mean and the patch_mean gives a scalar mean.
    ignore_closed_nodes : bool
        If True, do not incorporate values at closed nodes into the calc.
    return_components : bool
        If True, return a tuple, (array_of_magnitude,
        (array_of_slope_x_radians, array_of_slope_y_radians)).
        If false, return an array of floats of the slope magnitude.

    Returns
    -------
    float array or length-2 tuple of float arrays
        If return_components, returns (array_of_magnitude,
        (array_of_slope_x_radians, array_of_slope_y_radians)).
        If not return_components, returns an array of slope magnitudes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RadialModelGrid, RasterModelGrid
    >>> mg = RasterModelGrid((5, 5), 1.)
    >>> z = mg.node_x
    >>> slopes = mg.calc_slope_at_node(elevs=z)
    >>> np.allclose(slopes, np.pi / 4.)
    True
    >>> mg = RasterModelGrid((4, 5), 2.)
    >>> z = - mg.node_y
    >>> slope_mag, cmp = mg.calc_slope_at_node(elevs=z,
    ...                                        return_components=True)
    >>> np.allclose(slope_mag, np.pi / 4.)
    True
    >>> np.allclose(cmp[0], 0.)
    True
    >>> np.allclose(cmp[1], - np.pi / 4.)
    True
    >>> mg = RasterModelGrid((4, 4))
    >>> z = mg.node_x ** 2 + mg.node_y ** 2
    >>> slopes, cmp = mg.calc_slope_at_node(z, return_components=True)
    >>> slopes
    array([ 0.95531662,  1.10991779,  1.32082849,  1.37713803,  1.10991779,
            1.20591837,  1.3454815 ,  1.38904403,  1.32082849,  1.3454815 ,
            1.39288142,  1.41562833,  1.37713803,  1.38904403,  1.41562833,
            1.43030663])
    >>> np.allclose(cmp[0].reshape((4, 4))[:, 0],
    ...             cmp[1].reshape((4, 4))[0, :])  # test radial symmetry
    True

    LLCATS: NINF GRAD SURF
    """
    if method not in ('patch_mean', 'Horn'):
        raise ValueError('method name not understood')
    try:
        patches_at_node = grid.patches_at_node()
    except TypeError:  # was a property, not a fn (=> new style)
        if not ignore_closed_nodes:
            patches_at_node = np.ma.masked_where(
                grid.patches_at_node == -1, grid.patches_at_node,
                copy=False)
        else:
            patches_at_node = np.ma.masked_where(np.logical_not(
                grid.patches_present_at_node), grid.patches_at_node,
                copy=False)
    # now, we also want to mask any "closed" patches (any node closed)
    closed_patches = (grid.status_at_node[grid.nodes_at_patch] ==
                      CLOSED_BOUNDARY).sum(axis=1) > 0
    closed_patch_mask = np.logical_or(
        patches_at_node.mask, closed_patches[patches_at_node.data])

    if method == 'patch_mean':
        n_TR, n_TL, n_BL, n_BR = \
            grid.calc_unit_normals_at_patch_subtriangles(elevs)

        mean_slope_at_patches = grid.calc_slope_at_patch(
            elevs=elevs, ignore_closed_nodes=ignore_closed_nodes,
            subtriangle_unit_normals=(n_TR, n_TL, n_BL, n_BR))

        # now CAREFUL - patches_at_node is MASKED
        slopes_at_node_unmasked = mean_slope_at_patches[patches_at_node]
        slopes_at_node_masked = np.ma.array(slopes_at_node_unmasked,
                                            mask=closed_patch_mask)
        slope_mag = np.mean(slopes_at_node_masked, axis=1).data
        if return_components:
            (x_slope_patches, y_slope_patches) = grid.calc_grad_at_patch(
                elevs=elevs, ignore_closed_nodes=ignore_closed_nodes,
                subtriangle_unit_normals=(n_TR, n_TL, n_BL, n_BR),
                slope_magnitude=mean_slope_at_patches)
            x_slope_unmasked = x_slope_patches[patches_at_node]
            x_slope_masked = np.ma.array(x_slope_unmasked,
                                         mask=closed_patch_mask)
            x_slope = np.mean(x_slope_masked, axis=1).data
            y_slope_unmasked = y_slope_patches[patches_at_node]
            y_slope_masked = np.ma.array(y_slope_unmasked,
                                         mask=closed_patch_mask)
            y_slope = np.mean(y_slope_masked, axis=1).data
            mean_grad_x = x_slope
            mean_grad_y = y_slope
    elif method == 'Horn':
        z = np.empty(grid.number_of_nodes + 1, dtype=float)
        mean_grad_x = grid.empty(at='node', dtype=float)
        mean_grad_y = grid.empty(at='node', dtype=float)
        z[-1] = 0.
        try:
            z[:-1] = grid.at_node[elevs]
        except TypeError:
            z[:-1] = elevs
        # proof code for bad indexing:
        diags = grid.diagonal_neighbors_at_node.copy()  # LL order
        orthos = grid.adjacent_nodes_at_node.copy()
        # these have closed node neighbors...
        for dirs in (diags, orthos):
            dirs[dirs == BAD_INDEX_VALUE] = -1  # indexing to work
        # now make an array like patches_at_node to store the interim calcs
        patch_slopes_x = np.ma.zeros(patches_at_node.shape, dtype=float)
        patch_slopes_y = np.ma.zeros(patches_at_node.shape, dtype=float)
        diff_E = z[orthos[:, 0]] - z[:-1]
        diff_W = z[:-1] - z[orthos[:, 2]]
        diff_N = z[orthos[:, 1]] - z[:-1]
        diff_S = z[:-1] - z[orthos[:, 3]]
        patch_slopes_x[:, 0] = z[diags[:, 0]] - z[orthos[:, 1]] + diff_E
        patch_slopes_x[:, 1] = z[orthos[:, 1]] - z[diags[:, 1]] + diff_W
        patch_slopes_x[:, 2] = z[orthos[:, 3]] - z[diags[:, 2]] + diff_W
        patch_slopes_x[:, 3] = z[diags[:, 3]] - z[orthos[:, 3]] + diff_E
        patch_slopes_y[:, 0] = z[diags[:, 0]] - z[orthos[:, 0]] + diff_N
        patch_slopes_y[:, 1] = z[diags[:, 1]] - z[orthos[:, 2]] + diff_N
        patch_slopes_y[:, 2] = z[orthos[:, 2]] - z[diags[:, 2]] + diff_S
        patch_slopes_y[:, 3] = z[orthos[:, 0]] - z[diags[:, 3]] + diff_S
        patch_slopes_x /= (2. * grid.dx)
        patch_slopes_y /= (2. * grid.dy)
        patch_slopes_x.mask = closed_patch_mask
        patch_slopes_y.mask = closed_patch_mask
        mean_grad_x = patch_slopes_x.mean(axis=1).data
        mean_grad_y = patch_slopes_y.mean(axis=1).data
        slope_mag = np.arctan(np.sqrt(np.square(mean_grad_x) +
                                      np.square(mean_grad_y)))
        if return_components:
            mean_grad_x = np.arctan(mean_grad_x)
            mean_grad_y = np.arctan(mean_grad_y)

    if return_components:
        return slope_mag, (mean_grad_x, mean_grad_y)

    else:
        return slope_mag



import numpy as np
from numpy.testing import assert_array_equal

from landlab import FIXED_GRADIENT_BOUNDARY
from landlab import RasterModelGrid


def test_id_as_int():
    rmg = RasterModelGrid((4, 5))
    assert rmg.node_is_boundary(0)


def test_id_as_small_list():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.node_is_boundary([0]), np.array([True]))


def test_id_as_array():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.node_is_boundary(np.arange(20)),
        np.array([True,  True,  True,  True,  True,
                  True, False, False, False,  True,
                  True, False, False, False,  True,
                  True,  True,  True,  True,  True], dtype=bool))


def test_id_as_list():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.node_is_boundary([8, 9]), np.array([False, True]))


def test_boundary_flag():
    rmg = RasterModelGrid((4, 5))
    rmg.status_at_node[0] = FIXED_GRADIENT_BOUNDARY
    assert_array_equal(
        rmg.node_is_boundary(np.arange(20)),
        np.array([True,  True,  True,  True,  True,
                  True, False, False, False,  True,
                  True, False, False, False,  True,
                  True,  True,  True,  True,  True], dtype=bool))

    assert_array_equal(
        rmg.node_is_boundary(np.arange(20), boundary_flag=FIXED_GRADIENT_BOUNDARY),
        np.array([True, False, False, False, False,
                  False, False, False, False, False,
                  False, False, False, False, False,
                  False, False, False, False, False], dtype=bool))



import pytest

from landlab import RasterModelGrid
from landlab.components.flexure.flexure import Flexure
from landlab.components.flexure.flexure_1d import Flexure1D


@pytest.fixture
def flex():
    grid = RasterModelGrid((20, 20), spacing=10e3)
    return Flexure(grid)


@pytest.fixture
def flex1d():
    grid = RasterModelGrid((20, 20), spacing=10e3)
    return Flexure1D(grid)


FLEXURE_KEYWORDS = ("eet", "youngs", "rho_mantle", "rho_water", "gravity")


def pytest_generate_tests(metafunc):
    if "flexure_keyword" in metafunc.fixturenames:
        metafunc.parametrize("flexure_keyword", FLEXURE_KEYWORDS)



#!/usr/env/python

"""Landlab component that simulates landslide probability of failure as well
as mean relative wetness and probability of saturation.

Relative wetness and factor-of-safety are based on the infinite slope
stability model driven by topographic and soils inputs and recharge provided
by user as inputs to the component. For each node, component simulates mean
relative wetness as well as the probability of saturation based on Monte Carlo
simulation of relative wetness where the probability is the number of
iterations with relative wetness >= 1.0 divided by the number of iterations.
Probability of failure for each node is also simulated in the Monte Carlo
simulation as the number of iterations with factor-of-safety <= 1.0
divided by the number of iterations.

.. codeauthor:: R.Strauch, E.Istanbulluoglu, & S.S.Nudurupati
University of Washington

Ref 1: Strauch et. al. 2017, 'A hydro-climatological approach to predicting
regional landslide probability using Landlab, Earth Surface Dynamics, In prep.

Ref 2: 'The Landlab LandslideProbability Component User Manual' @
https://github.com/RondaStrauch/pub_strauch_etal_esurf/blob/master/LandslideComponentUsersManual.docx

Created on Thu Aug 20, 2015
Last edit June 7, 2017
"""

# %% Import Libraries
from landlab import Component
from landlab.utils.decorators import use_file_name_or_kwds
import numpy as np
import scipy.constants
from scipy import interpolate
from statsmodels.distributions.empirical_distribution import ECDF
import copy

# %% Instantiate Object


class LandslideProbability(Component):
    """Landslide probability component using the infinite slope stability
    model.
    
    Landlab component designed to calculate probability of failure at
    each grid node based on the infinite slope stability model
    stability index (Factor of Safety).

    The driving force for failure is provided by the user in the form of
    groundwater recharge; four options for providing recharge are supported.
    The model uses topographic and soil characteristics provided as input
    by the user.

    The main method of the LandslideProbability class is
    calculate_landslide_probability(), which calculates the mean soil relative
    wetness, probability of soil saturation, and probability of failure at
    each node based on a Monte Carlo simulation.

    **Usage:**

    Option 1 - Uniform recharge::

        LandslideProbability(grid, number_of_iterations=250,
                             groundwater__recharge_distribution='uniform', 
                             groundwater__recharge_min_value=5.,
                             groundwater__recharge_max_value=121.)

    Option 2 - Lognormal recharge::

        LandslideProbability(grid, number_of_iterations=250,
                             groundwater__recharge_distribution='lognormal',
                             groundwater__recharge_mean=30.,
                             groundwater__recharge_standard_deviation=0.25)

    Option 3 - Lognormal_spatial recharge::

        LandslideProbability(grid, number_of_iterations=250,
                             groundwater__recharge_distribution='lognormal_spatial',
                             groundwater__recharge_mean=np.random.randint(20, 120, grid_size),
                             groundwater__recharge_standard_deviation=np.random.rand(grid_size))

    Option 4 - Data_driven_spatial recharge::

        LandslideProbability(grid, number_of_iterations=250,
                             groundwater__recharge_distribution='data_driven_spatial',
                             groundwater__recharge_HSD_inputs=[HSD_dict, HSD_id_dict,
                             fract_dict])

    Examples
    ----------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.landslides import LandslideProbability
    >>> import numpy as np
    
    Create a grid on which to calculate landslide probability.
    
    >>> grid = RasterModelGrid((5, 4), spacing=(0.2, 0.2))
    
    Check the number of core nodes.
    
    >>> grid.number_of_core_nodes
    6
    
    The grid will need some input data. To check the names of the fields
    that provide the input to this component, use the *input_var_names*
    class property.
    
    >>> sorted(LandslideProbability.input_var_names)  # doctest: +NORMALIZE_WHITESPACE
    ['soil__density',
     'soil__internal_friction_angle',
     'soil__maximum_total_cohesion',
     'soil__minimum_total_cohesion',
     'soil__mode_total_cohesion',
     'soil__saturated_hydraulic_conductivity',
     'soil__thickness',
     'soil__transmissivity',
     'topographic__slope',
     'topographic__specific_contributing_area']
    
    Check the units for the fields.
    
    >>> LandslideProbability.var_units('topographic__specific_contributing_area')
    'm'
    
    Create an input field.
    
    >>> grid.at_node['topographic__slope'] = np.random.rand(grid.number_of_nodes)
    
    If you are not sure about one of the input or output variables, you can
    get help for specific variables.
    
    >>> LandslideProbability.var_help('soil__transmissivity')  # doctest: +NORMALIZE_WHITESPACE
    name: soil__transmissivity
    description:
      mode rate of water transmitted through a unit width of saturated
      soil - either provided or calculated with Ksat and soil depth
    units: m2/day
    at: node
    intent: in
    
    Additional required fields for component.
    
    >>> scatter_dat = np.random.randint(1, 10, grid.number_of_nodes)
    >>> grid.at_node['topographic__specific_contributing_area'] = np.sort(
    ...      np.random.randint(30, 900, grid.number_of_nodes))
    >>> grid.at_node['soil__transmissivity'] = np.sort(
    ...      np.random.randint(5, 20, grid.number_of_nodes), -1)
    >>> grid.at_node['soil__saturated_hydraulic_conductivity'] = np.sort(
    ...      np.random.randint(2, 10, grid.number_of_nodes), -1) 
    >>> grid.at_node['soil__mode_total_cohesion'] = np.sort(
    ...      np.random.randint(30, 900, grid.number_of_nodes))
    >>> grid.at_node['soil__minimum_total_cohesion'] = (
    ...      grid.at_node['soil__mode_total_cohesion'] - scatter_dat)
    >>> grid.at_node['soil__maximum_total_cohesion'] = (
    ...      grid.at_node['soil__mode_total_cohesion'] + scatter_dat)
    >>> grid.at_node['soil__internal_friction_angle'] = np.sort(
    ...      np.random.randint(26, 40, grid.number_of_nodes))
    >>> grid.at_node['soil__thickness'] = np.sort(
    ...      np.random.randint(1, 10, grid.number_of_nodes))
    >>> grid.at_node['soil__density'] = (2000. * np.ones(grid.number_of_nodes))
    
    Instantiate the 'LandslideProbability' component to work on this grid,
    and run it.
    
    >>> ls_prob = LandslideProbability(grid)
    >>> np.allclose(grid.at_node['landslide__probability_of_failure'], 0.)
    True
    
    Run the *calculate_landslide_probability* method to update output
    variables with grid
    
    >>> ls_prob.calculate_landslide_probability()
    
    Check the output variable names.
    
    >>> sorted(ls_prob.output_var_names) # doctest: +NORMALIZE_WHITESPACE
    ['landslide__probability_of_failure',
     'soil__mean_relative_wetness',
     'soil__probability_of_saturation']
    
    Check the output from the component, including array at one node.
    
    >>> np.allclose(grid.at_node['landslide__probability_of_failure'], 0.)
    False
    >>> core_nodes = ls_prob.grid.core_nodes
    """

    # component name
    _name = 'Landslide Probability'
    __version__ = '1.0'
    # component requires these values to do its calculation, get from driver
    _input_var_names = (
        'topographic__specific_contributing_area',
        'topographic__slope',
        'soil__transmissivity',
        'soil__saturated_hydraulic_conductivity',
        'soil__mode_total_cohesion',
        'soil__minimum_total_cohesion',
        'soil__maximum_total_cohesion',
        'soil__internal_friction_angle',
        'soil__density',
        'soil__thickness',
        )

    #  component creates these output values
    _output_var_names = (
        'soil__mean_relative_wetness',
        'landslide__probability_of_failure',
        'soil__probability_of_saturation',
        )

    # units for each parameter and output
    _var_units = {
        'topographic__specific_contributing_area': 'm',
        'topographic__slope': 'tan theta',
        'soil__transmissivity': 'm2/day',
        'soil__saturated_hydraulic_conductivity': 'm/day',
        'soil__mode_total_cohesion': 'Pa or kg/m-s2',
        'soil__minimum_total_cohesion': 'Pa or kg/m-s2',
        'soil__maximum_total_cohesion': 'Pa or kg/m-s2',
        'soil__internal_friction_angle': 'degrees',
        'soil__density': 'kg/m3',
        'soil__thickness': 'm',
        'soil__mean_relative_wetness': 'None',
        'landslide__probability_of_failure': 'None',
        'soil__probability_of_saturation': 'None',
        }

    # grid centering of each field and variable
    _var_mapping = {
        'topographic__specific_contributing_area': 'node',
        'topographic__slope': 'node',
        'soil__transmissivity': 'node',
        'soil__saturated_hydraulic_conductivity': 'node',
        'soil__mode_total_cohesion': 'node',
        'soil__minimum_total_cohesion': 'node',
        'soil__maximum_total_cohesion': 'node',
        'soil__internal_friction_angle': 'node',
        'soil__density': 'node',
        'soil__thickness': 'node',
        'soil__mean_relative_wetness': 'node',
        'landslide__probability_of_failure': 'node',
        'soil__probability_of_saturation': 'node',
        }

    # short description of each field
    _var_doc = {
        'topographic__specific_contributing_area':
            ('specific contributing (upslope area/cell face )' +
             ' that drains to node'),
        'topographic__slope':
        'slope of surface at node represented by tan theta',
        'soil__transmissivity':
            ('mode rate of water transmitted' +
             ' through a unit width of saturated soil - ' +
              'either provided or calculated with Ksat ' + 
               'and soil depth'),
        'soil__saturated_hydraulic_conductivity':
            ('mode rate of water transmitted' +
             ' through soil - provided if transmissivity ' +
              'is NOT provided to calculate tranmissivity ' + 
               ' with soil depth'),
        'soil__mode_total_cohesion':
            'mode of combined root and soil cohesion at node',
        'soil__minimum_total_cohesion':
            'minimum of combined root and soil cohesion at node',
        'soil__maximum_total_cohesion':
            'maximum of combined root and soil cohesion at node',
        'soil__internal_friction_angle':
            ('critical angle just before failure' +
             ' due to friction between particles'),
        'soil__density': 'wet bulk density of soil',
        'soil__thickness': 'soil depth to restrictive layer',
        'soil__mean_relative_wetness':
            ('Indicator of soil wetness;' +
             ' relative depth perched water table' +
             ' within the soil layer'),
        'landslide__probability_of_failure':
            ('number of times FS is <=1 out of number of' +
             ' iterations user selected'),
        'soil__probability_of_saturation':
            ('number of times relative wetness is >=1 out of' +
             ' number of iterations user selected'),
        }

    # Run Component
    @use_file_name_or_kwds
    def __init__(self, grid, number_of_iterations=250,
                 groundwater__recharge_distribution='uniform',
                 groundwater__recharge_min_value=20.,
                 groundwater__recharge_max_value=120.,
                 groundwater__recharge_mean=None,
                 groundwater__recharge_standard_deviation=None,
                 groundwater__recharge_HSD_inputs=[],
                 seed=0, **kwds):
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A raster grid.
        number_of_iterations: int, optional
            Number of iterations to run Monte Carlo simulation (default=250).
        groundwater__recharge_distribution: str, optional
            single word indicating recharge distribution, either 'uniform',
            'lognormal', 'lognormal_spatial,' or 'data_driven_spatial'.
             (default='uniform')
        groundwater__recharge_min_value: float, optional (mm/d)
            minium groundwater recharge for 'uniform' (default=20.)
        groundwater__recharge_max_value: float, optional (mm/d)
            maximum groundwater recharge for 'uniform' (default=120.)
        groundwater__recharge_mean: float, optional (mm/d) 
            mean grounwater recharge for 'lognormal'
            and 'lognormal_spatial' (default=None)
        groundwater__recharge_standard_deviation: float, optional (mm/d)
            standard deviation of grounwater recharge for 'lognormal'
            and 'lognormal_spatial' (default=None)
        groundwater__recharge_HSD_inputs: list, optional
            list of 3 dictionaries in order (default=[]) - HSD_dict
            {Hydrologic Source Domain (HSD) keys: recharge numpy array values},
            {node IDs keys: list of HSD_Id values}, HSD_fractions {node IDS
            keys: list of HSD fractions values} (none) 
            Note: this input method is a very specific one, and to use this method,
            one has to refer Ref 1 & Ref 2 mentioned above, as this set of
            inputs require rigorous pre-processing of data.
        g: float, optional (m/sec^2)
            acceleration due to gravity.
        seed: int, optional
            seed for random number generation. if seed is assigned any value
            other than the default value of zero, it will create different
            sequence. To create a certain sequence repititively, use the same
            value as input for seed.
        """
        # Initialize seeded random number generation        
        self._seed_generator(seed)

        super(LandslideProbability, self).__init__(grid)

        # Store grid and parameters and do unit conversions
        self.n = int(number_of_iterations)
        self._g = kwds.get('g', scipy.constants.g)
        self.groundwater__recharge_distribution = (
            groundwater__recharge_distribution)
        # Following code will deal with the input distribution and associated
        # parameters
        # Uniform distribution
        if self.groundwater__recharge_distribution == 'uniform':
            self._recharge_min = groundwater__recharge_min_value
            self._recharge_max = groundwater__recharge_max_value
            self._Re = np.random.uniform(self._recharge_min,
                                         self._recharge_max,
                                         size=self.n)
            self._Re /= 1000. # Convert mm to m
        # Lognormal Distribution - Uniform in space
        elif self.groundwater__recharge_distribution == 'lognormal':
            assert (groundwater__recharge_mean != None), (
                'Input mean of the distribution!')
            assert (groundwater__recharge_standard_deviation != None), (
                'Input standard deviation of the distribution!')
            self._recharge_mean = groundwater__recharge_mean
            self._recharge_stdev = groundwater__recharge_standard_deviation
            self._mu_lognormal = np.log((self._recharge_mean**2)/np.sqrt(
                self._recharge_stdev**2 + self._recharge_mean**2))
            self._sigma_lognormal = np.sqrt(np.log((self._recharge_stdev**2)/(
                self._recharge_mean**2)+1))
            self._Re = np.random.lognormal(self._mu_lognormal,
                                          self._sigma_lognormal, self.n)
            self._Re /= 1000. # Convert mm to m
        # Lognormal Distribution - Variable in space                                  
        elif self.groundwater__recharge_distribution == 'lognormal_spatial':
            assert (groundwater__recharge_mean.shape[0] == (
                self.grid.number_of_nodes)), (
                'Input array should be of the length of grid.number_of_nodes!')
            assert (groundwater__recharge_standard_deviation.shape[0] == (
                self.grid.number_of_nodes)), (
                'Input array should be of the length of grid.number_of_nodes!')
            self._recharge_mean = groundwater__recharge_mean
            self._recharge_stdev = groundwater__recharge_standard_deviation
        # Custom HSD inputs - Hydrologic Source Domain -> Model Domain
        elif self.groundwater__recharge_distribution == 'data_driven_spatial':
            self._HSD_dict = groundwater__recharge_HSD_inputs[0]
            self._HSD_id_dict = groundwater__recharge_HSD_inputs[1]
            self._fract_dict = groundwater__recharge_HSD_inputs[2]
            self._interpolate_HSD_dict()

        # Check if all input fields are initialized
        for name in self._input_var_names:
            if name not in self.grid.at_node:
                self.grid.add_zeros(name, at=self._var_mapping[name],
                                    units=self._var_units[name])

        # Check if all output fields are initialized
        for name in self._output_var_names:
            if name not in self.grid.at_node:
                self.grid.add_zeros(name, at=self._var_mapping[name],
                                    units=self._var_units[name])

        # Create a switch to imply whether Ksat is provided.
        if np.all(
            self.grid.at_node['soil__saturated_hydraulic_conductivity'] == 0):
            self.Ksat_provided = 0  # False
        else:
            self.Ksat_provided = 1  # True

        self._nodal_values = self.grid.at_node


    def calculate_factor_of_safety(self, i):
        """Method to calculate factor of safety.

        Method calculates factor-of-safety stability index by using
        node specific parameters, creating distributions of these parameters,
        and calculating the index by sampling these distributions 'n' times.

        The index is calculated from the 'infinite slope stabilty
        factor-of-safety equation' in the format of Pack RT, Tarboton DG,
        and Goodwin CN (1998),The SINMAP approach to terrain stability mapping.

        Parameters
        ----------
        i: int
            index of core node ID.
        """
        # generate distributions to sample from to provide input parameters
        # currently triangle distribution using mode, min, & max
        self._a = np.float32(self.grid.at_node[
            'topographic__specific_contributing_area'][i])
        self._theta = np.float32(self.grid.at_node['topographic__slope'][i])
        self._Tmode = np.float32(self.grid.at_node['soil__transmissivity'][i])
        self._Ksatmode = np.float32(self.grid.at_node[
            'soil__saturated_hydraulic_conductivity'][i])
        self._Cmode = np.float32(
                self.grid.at_node['soil__mode_total_cohesion'][i])
        self._Cmin = np.float32(
                self.grid.at_node['soil__minimum_total_cohesion'][i])
        self._Cmax = np.float32(
                self.grid.at_node['soil__maximum_total_cohesion'][i])
        self._phi_mode = np.float32(
                self.grid.at_node['soil__internal_friction_angle'][i])
        self._rho = np.float32(self.grid.at_node['soil__density'][i])
        self._hs_mode = np.float32(self.grid.at_node['soil__thickness'][i])

        # recharge distribution based on distribution type
        if self.groundwater__recharge_distribution == 'data_driven_spatial':
            self._calculate_HSD_recharge(i)
            self._Re /= 1000.  # mm->m
        elif self.groundwater__recharge_distribution == 'lognormal_spatial':
            mu_lognormal = np.log((self._recharge_mean[i]**2)/np.sqrt(
                self._recharge_stdev[i]**2 + self._recharge_mean[i]**2))
            sigma_lognormal = np.sqrt(np.log((self._recharge_stdev[i]**2)/(
                self._recharge_mean[i]**2)+1))
            self._Re = np.random.lognormal(mu_lognormal,
                                          sigma_lognormal, self.n)
            self._Re /= 1000. # Convert mm to m

        # Cohesion
        # if don't provide fields of min and max C, uncomment 2 lines below
        #    Cmin = self._Cmode-0.3*self._Cmode
        #    Cmax = self._Cmode+0.3*self._Cmode
        self._C = np.random.triangular(self._Cmin, self._Cmode,
                                      self._Cmax, size=self.n)

        # phi - internal angle of friction provided in degrees
        phi_min = self._phi_mode-0.18*self._phi_mode
        phi_max = self._phi_mode+0.32*self._phi_mode
        self._phi = np.random.triangular(phi_min, self._phi_mode,
                                        phi_max, size=self.n)
        # soil thickness
        #hs_min = min(0.005, self._hs_mode-0.3*self._hs_mode) # Alternative
        hs_min = self._hs_mode-0.3*self._hs_mode 
        hs_max = self._hs_mode+0.1*self._hs_mode
        self._hs = np.random.triangular(hs_min, self._hs_mode,
                                       hs_max, size=self.n)
        self._hs[self._hs <= 0.] = 0.005
        if self.Ksat_provided:
            # Hydraulic conductivity (Ksat)
            Ksatmin = self._Ksatmode-(0.3*self._Ksatmode)
            Ksatmax = self._Ksatmode+(0.1*self._Ksatmode)
            self._Ksat = np.random.triangular(Ksatmin, self._Ksatmode,
                                              Ksatmax, size=self.n)
            self._T = self._Ksat*self._hs
        else:
            # Transmissivity (T)
            Tmin = self._Tmode-(0.3*self._Tmode)
            Tmax = self._Tmode+(0.1*self._Tmode)
            self._T = np.random.triangular(Tmin, self._Tmode,
                                           Tmax, size=self.n)

        # calculate Factor of Safety for n number of times
        # calculate components of FS equation
        self._C_dim = self._C/(self._hs*self._rho*self._g)  # dimensionless cohesion
        self._rel_wetness = ((self._Re)/self._T)*(self._a/np.sin(
            np.arctan(self._theta)))                    # relative wetness
        # calculate probability of saturation
        countr = 0
        for val in self._rel_wetness:            # find how many RW values >= 1
            if val >= 1.0:
                countr = countr + 1  # number with RW values (>=1)
        # probability: No. high RW values/total No. of values (n)
        self._soil__probability_of_saturation = np.float32(countr)/self.n
        # Maximum Rel_wetness = 1.0        
        np.place(self._rel_wetness, self._rel_wetness > 1, 1.0)
        self._soil__mean_relative_wetness = np.mean(self._rel_wetness)
        Y = np.tan(np.radians(self._phi))*(1 - (self._rel_wetness*0.5))
        # convert from degrees; 0.5 = water to soil density ratio
        # calculate Factor-of-safety
        self._FS = (self._C_dim/np.sin(np.arctan(self._theta))) + (
            np.cos(np.arctan(self._theta)) *
            (Y/np.sin(np.arctan(self._theta))))
        count = 0
        for val in self._FS:                   # find how many FS values <= 1
            if val <= 1.0:
                count = count + 1   # number with unstable FS values (<=1)
        # probability: No. unstable values/total No. of values (n)
        self._landslide__probability_of_failure = np.float32(count)/self.n


    def calculate_landslide_probability(self, **kwds):
        """Main method of Landslide Probability class.

        Method creates arrays for output variables then loops through all
        the core nodes to run the method 'calculate_factor_of_safety.'
        Output parameters probability of failure, mean relative wetness,
        and probability of saturation are assigned as fields to nodes. 
        """
        # Create arrays for data with -9999 as default to store output
        self.mean_Relative_Wetness = np.full(self.grid.number_of_nodes,
                                             -9999.)
        self.prob_fail = np.full(self.grid.number_of_nodes, -9999.)
        self.prob_sat = np.full(self.grid.number_of_nodes, -9999.) 
        # Run factor of safety Monte Carlo for all core nodes in domain
        # i refers to each core node id
        for i in self.grid.core_nodes:
            self.calculate_factor_of_safety(i)
            # Populate storage arrays with calculated values
            self.mean_Relative_Wetness[i] = self._soil__mean_relative_wetness
            self.prob_fail[i] = self._landslide__probability_of_failure
            self.prob_sat[i] = self._soil__probability_of_saturation
        # Values can't be negative
        self.mean_Relative_Wetness[
            self.mean_Relative_Wetness < 0.] = 0.
        self.prob_fail[self.prob_fail < 0.] = 0.
        # assign output fields to nodes
        self.grid.at_node['soil__mean_relative_wetness'] = (
            self.mean_Relative_Wetness)
        self.grid.at_node['landslide__probability_of_failure'] = (
                self.prob_fail)
        self.grid.at_node['soil__probability_of_saturation'] = self.prob_sat


    def _seed_generator(self, seed=0):
        """Method to initiate random seed.
        
        Seed the random-number generator. This method will create the same
        sequence again by re-seeding with the same value (default value is
        zero). To create a sequence other than the default, assign non-zero
        value for seed.
        """
        np.random.seed(seed)


    def _interpolate_HSD_dict(self):
        """Method to extrapolate input data.
        
        This method uses a non-parametric approach to expand the input
        recharge array to the length of number of iterations. Output is
        a new dictionary of interpolated recharge for each HSD id.
        """
        HSD_dict = copy.deepcopy(self._HSD_dict)
        # First generate interpolated Re for each HSD grid
        Yrand = np.sort(np.random.rand(self.n))
        # n random numbers (0 to 1) in a column
        for vkey in HSD_dict.keys():
            if isinstance(HSD_dict[vkey], int):
                continue       # loop back up if value is integer (e.g. -9999)
            Re_temp = HSD_dict[vkey]	 # an array of annual Re for 1 HSD grid
            Fx = ECDF(Re_temp)  # instantiate to get probabilities with Re
            Fx_ = Fx(Re_temp)    # probability array associated with Re data
            # interpolate function based on recharge data & probability
            f = interpolate.interp1d(Fx_, Re_temp, bounds_error=False,
                                     fill_value=min(Re_temp))
            # array of Re interpolated from Yrand probabilities (n count)
            Re_interpolated = f(Yrand)
            # replace values in HSD_dict with interpolated Re
            HSD_dict[vkey] = Re_interpolated

        self._interpolated_HSD_dict = HSD_dict


    def _calculate_HSD_recharge(self, i):
        """Method to calculate recharge based on upstream fractions.
        
        This method calculates the resultant recharge at node i of the
        model domain, using recharge of contributing HSD ids and the areal
        fractions of upstream contributing HSD ids. Output is a numpy array
        of recharge at node i.        
        """
        store_Re = np.zeros(self.n)
        HSD_id_list = self._HSD_id_dict[i]
        fract_list = self._fract_dict[i]
        for j in range(0, len(HSD_id_list)):
            Re_temp = self._interpolated_HSD_dict[HSD_id_list[j]]
            fract_temp = fract_list[j]
            Re_adj = (Re_temp*fract_temp)
            store_Re = np.vstack((store_Re, np.array(Re_adj)))
        self._Re = np.sum(store_Re, 0)



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.grid import raster_funcs as rfuncs


def test_face_with_scalar_cell_id():
    """Test using a scalar arg for cell."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0]), 0)
    assert_array_equal(node_ids, np.array([[1]]))


def test_face_with_iterable_cell_id():
    """Test using iterable arg for cell."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0]), (5, ))
    assert_array_equal(node_ids, np.array([[8]]))


def test_face_with_no_cell_id():
    """Test without an arg for cells to mean all cells."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0]))
    assert_array_equal(node_ids, np.array([[1, 2, 3, 6, 7, 8]]).T)


def test_face_multiple_faces():
    """Test getting nodes for more than one corner."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0, 1]), (4, ))
    assert_array_equal(node_ids, np.array([[7, 11]]))


def test_face_multiple_corners_and_cells():
    """Test getting nodes for more than one corner and more than one cell."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0, 1]), (4, 5))
    assert_array_equal(node_ids, np.array([[7, 11], [8, 12]]))


def test_face_type_tuple():
    """Test using tuple as face arg."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    node_ids = rfuncs.neighbor_node_at_cell(rmg, (0, 1), (4, 5))
    assert_array_equal(node_ids, np.array([[7, 11], [8, 12]]))


def test_face_type_list():
    """Test using list as face arg."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    node_ids = rfuncs.neighbor_node_at_cell(rmg, [0, 1], (4, 5))
    assert_array_equal(node_ids, np.array([[7, 11], [8, 12]]))


def test_face_type_scalar():
    """Test using scalar as face arg."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    node_ids = rfuncs.neighbor_node_at_cell(rmg, 2, (4, ))
    assert_array_equal(node_ids, np.array([17]))

    node_ids = rfuncs.neighbor_node_at_cell(rmg, 2, (4, 5))
    assert_array_equal(node_ids, np.array([17, 18]))

    node_ids = rfuncs.neighbor_node_at_cell(rmg, 1)
    assert_array_equal(node_ids, np.array([5, 6, 7, 10, 11, 12]))



"""Test the flow accumulator component.

@author: krb
"""
# Created on Thurs Nov 12, 2015
import os

import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

import landlab
from landlab import RasterModelGrid, HexModelGrid, RadialModelGrid, FieldError
from landlab.components.flow_accum import FlowAccumulator

from landlab.components.flow_routing.lake_mapper import DepressionFinderAndRouter

from landlab.components.flow_director import(FlowDirectorD8,
                                             FlowDirectorDINF,
                                             FlowDirectorMFD,
                                             FlowDirectorSteepest)

from landlab import CLOSED_BOUNDARY
from landlab import BAD_INDEX_VALUE as XX

_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def test_check_fields():
    """Check to make sure the right fields have been created."""

    mg = RasterModelGrid((10,10), spacing=(1, 1))
    z = mg.add_field('topographic__elevation', mg.node_x**2 + mg.node_y**2, at = 'node')

    fa = FlowAccumulator(mg)
    assert_array_equal(z, mg.at_node['topographic__elevation'])
    assert_array_equal(np.zeros(100), mg.at_node['drainage_area'])
    assert_array_equal(np.ones(100), mg.at_node['water__unit_flux_in'])

    fa = FlowAccumulator(mg, runoff_rate=2.)
    assert_array_equal(np.full(100, 2.), mg.at_node['water__unit_flux_in'])


def test_director_adding_methods_are_equivalent_Steepest():
    """Check that different methods to specifying the director are the same."""

    mg0 = RasterModelGrid((10,10), spacing=(1, 1))
    z0 = mg0.add_field('topographic__elevation', mg0.node_x**2 + mg0.node_y**2, at = 'node')
    fa0 = FlowAccumulator(mg0, flow_director='D4')
    fa0.run_one_step()

    mg1 = RasterModelGrid((10,10), spacing=(1, 1))
    z1 = mg1.add_field('topographic__elevation', mg1.node_x**2 + mg1.node_y**2, at = 'node')
    fa1 = FlowAccumulator(mg1, flow_director='Steepest')
    fa1.run_one_step()

    mg2 = RasterModelGrid((10,10), spacing=(1, 1))
    z2 = mg2.add_field('topographic__elevation', mg2.node_x**2 + mg2.node_y**2, at = 'node')
    fa2 = FlowAccumulator(mg2, flow_director=FlowDirectorSteepest)
    fa2.run_one_step()

    mg3 = RasterModelGrid((10,10), spacing=(1, 1))
    z3 = mg3.add_field('topographic__elevation', mg3.node_x**2 + mg3.node_y**2, at = 'node')
    fd = FlowDirectorSteepest(mg3)
    fa3 = FlowAccumulator(mg3, flow_director=fd)
    fa3.run_one_step()

    for key in mg0.at_node.keys():
        assert_array_equal(mg0.at_node[key],
                           mg1.at_node[key])

        assert_array_equal(mg1.at_node[key],
                           mg2.at_node[key])

        assert_array_equal(mg2.at_node[key],
                           mg3.at_node[key])


def test_director_adding_methods_are_equivalent_D8():
    """Check that different methods to specifying the director are the same."""

    mg0 = RasterModelGrid((10,10), spacing=(1, 1))
    z0 = mg0.add_field('topographic__elevation', mg0.node_x**2 + mg0.node_y**2, at = 'node')
    fa0 = FlowAccumulator(mg0, flow_director='D8')
    fa0.run_one_step()

    mg1 = RasterModelGrid((10,10), spacing=(1, 1))
    z1 = mg1.add_field('topographic__elevation', mg1.node_x**2 + mg1.node_y**2, at = 'node')
    fa1 = FlowAccumulator(mg1, flow_director='FlowDirectorD8')
    fa1.run_one_step()

    mg2 = RasterModelGrid((10,10), spacing=(1, 1))
    z2 = mg2.add_field('topographic__elevation', mg2.node_x**2 + mg2.node_y**2, at = 'node')
    fa2 = FlowAccumulator(mg2, flow_director=FlowDirectorD8)
    fa2.run_one_step()

    mg3 = RasterModelGrid((10,10), spacing=(1, 1))
    z3 = mg3.add_field('topographic__elevation', mg3.node_x**2 + mg3.node_y**2, at = 'node')
    fd = FlowDirectorD8(mg3)
    fa3 = FlowAccumulator(mg3, flow_director=fd)
    fa3.run_one_step()

    for key in mg0.at_node.keys():
        assert_array_equal(mg0.at_node[key],
                           mg1.at_node[key])

        assert_array_equal(mg1.at_node[key],
                           mg2.at_node[key])

        assert_array_equal(mg2.at_node[key],
                           mg3.at_node[key])


def test_director_adding_methods_are_equivalent_Dinf():
    """Check that different methods to specifying the director are the same."""

    mg0 = RasterModelGrid((10,10), spacing=(1, 1))
    z0 = mg0.add_field('topographic__elevation', mg0.node_x**2 + mg0.node_y**2, at = 'node')
    fa0 = FlowAccumulator(mg0, flow_director='DINF')
    fa0.run_one_step()

    mg1 = RasterModelGrid((10,10), spacing=(1, 1))
    z1 = mg1.add_field('topographic__elevation', mg1.node_x**2 + mg1.node_y**2, at = 'node')
    fa1 = FlowAccumulator(mg1, flow_director='FlowDirectorDINF')
    fa1.run_one_step()

    mg2 = RasterModelGrid((10,10), spacing=(1, 1))
    z2 = mg2.add_field('topographic__elevation', mg2.node_x**2 + mg2.node_y**2, at = 'node')
    fa2 = FlowAccumulator(mg2, flow_director=FlowDirectorDINF)
    fa2.run_one_step()

    mg3 = RasterModelGrid((10,10), spacing=(1, 1))
    z3 = mg3.add_field('topographic__elevation', mg3.node_x**2 + mg3.node_y**2, at = 'node')
    fd = FlowDirectorDINF(mg3)
    fa3 = FlowAccumulator(mg3, flow_director=fd)
    fa3.run_one_step()

    for key in mg0.at_node.keys():
        assert_array_equal(mg0.at_node[key],
                           mg1.at_node[key])

        assert_array_equal(mg1.at_node[key],
                           mg2.at_node[key])

        assert_array_equal(mg2.at_node[key],
                           mg3.at_node[key])


def test_director_adding_methods_are_equivalent_MFD():
    """Check that different methods to specifying the director are the same."""

    mg0 = RasterModelGrid((10,10), spacing=(1, 1))
    z0 = mg0.add_field('topographic__elevation', mg0.node_x**2 + mg0.node_y**2, at = 'node')
    fa0 = FlowAccumulator(mg0, flow_director='MFD')
    fa0.run_one_step()

    mg1 = RasterModelGrid((10,10), spacing=(1, 1))
    z1 = mg1.add_field('topographic__elevation', mg1.node_x**2 + mg1.node_y**2, at = 'node')
    fa1 = FlowAccumulator(mg1, flow_director='FlowDirectorMFD')
    fa1.run_one_step()

    mg2 = RasterModelGrid((10,10), spacing=(1, 1))
    z2 = mg2.add_field('topographic__elevation', mg2.node_x**2 + mg2.node_y**2, at = 'node')
    fa2 = FlowAccumulator(mg2, flow_director=FlowDirectorMFD)
    fa2.run_one_step()

    mg3 = RasterModelGrid((10,10), spacing=(1, 1))
    z3 = mg3.add_field('topographic__elevation', mg3.node_x**2 + mg3.node_y**2, at = 'node')
    fd = FlowDirectorMFD(mg3)
    fa3 = FlowAccumulator(mg3, flow_director=fd)
    fa3.run_one_step()

    for key in mg0.at_node.keys():
        assert_array_equal(mg0.at_node[key],
                           mg1.at_node[key])

        assert_array_equal(mg1.at_node[key],
                           mg2.at_node[key])

        assert_array_equal(mg2.at_node[key],
                           mg3.at_node[key])


def test_passing_a_bad_component():
    """Check that a random component can't be a director."""
    from landlab.components import ChiFinder

    mg = RasterModelGrid((10,10), spacing=(1, 1))
    _ = mg.add_field('topographic__elevation', mg.node_x + mg.node_y, at = 'node')
    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    with pytest.raises(ValueError):
        FlowAccumulator(mg, 'topographic__elevation', flow_director=ChiFinder)


def test_error_for_to_many_with_depression():
    """Check that an error is thrown when to_many methods started DF."""

    mg0 = RasterModelGrid((10,10), spacing=(1, 1))
    z0 = mg0.add_field('topographic__elevation', mg0.node_x**2 + mg0.node_y**2, at = 'node')

    mg1 = RasterModelGrid((10,10), spacing=(1, 1))
    z1 = mg1.add_field('topographic__elevation', mg1.node_x**2 + mg1.node_y**2, at = 'node')


    with pytest.raises(ValueError):
        FlowAccumulator(mg0, flow_director='MFD', depression_finder='DepressionFinderAndRouter')
    with pytest.raises(ValueError):
        FlowAccumulator(mg0, flow_director='DINF', depression_finder='DepressionFinderAndRouter')

    fa0 = FlowAccumulator(mg0, flow_director='MFD')
    fa0.run_one_step()
    with pytest.raises(ValueError):
        DepressionFinderAndRouter(mg0)

    fa1 = FlowAccumulator(mg1, flow_director='DINF')
    fa1.run_one_step()
    with pytest.raises(ValueError):
        DepressionFinderAndRouter(mg1)


def test_fields():
    """Check to make sure the right fields have been created.

    Check that the sizes are also correct.
    """
    mg = RasterModelGrid((10,10), spacing=(1, 1))
    _ = mg.add_field('topographic__elevation', mg.node_x + mg.node_y, at = 'node')
    fa = FlowAccumulator(mg)
    fa.run_one_step()

    assert sorted(list(mg.at_node.keys())) == ['drainage_area',
                                               'flow__data_structure_delta',
                                               'flow__link_to_receiver_node',
                                               'flow__receiver_node',
                                               'flow__sink_flag',
                                               'flow__upstream_node_order',
                                               'surface_water__discharge',
                                               'topographic__elevation',
                                               'topographic__steepest_slope',
                                               'water__unit_flux_in']
    assert sorted(list(mg.at_link.keys())) == ['flow__data_structure_D']

    mg2 = RasterModelGrid((10,10), spacing=(1, 1))
    _ = mg2.add_field('topographic__elevation', mg2.node_x + mg2.node_y, at = 'node')
    fa2 = FlowAccumulator(mg2, flow_director='MFD')
    fa2.run_one_step()
    assert sorted(list(mg2.at_node.keys())) == ['drainage_area',
                                                'flow__data_structure_delta',
                                                'flow__link_to_receiver_node',
                                                'flow__links_to_receiver_nodes',
                                                'flow__receiver_node',
                                                'flow__receiver_nodes',
                                                'flow__receiver_proportions',
                                                'flow__sink_flag',
                                                'flow__upstream_node_order',
                                                'surface_water__discharge',
                                                'topographic__elevation',
                                                'topographic__steepest_slope',
                                                'water__unit_flux_in']

    assert sorted(list(mg2.at_link.keys())) == ['flow__data_structure_D']


def test_accumulated_area_closes():
    """Check that accumulated area is area of core nodes."""

    fds= ['Steepest','D8','MFD','DINF']

    for fd in fds:
        mg = RasterModelGrid((10,10), spacing=(1, 1))
        z = mg.add_field('topographic__elevation', mg.node_x + mg.node_y, at = 'node')
        fa = FlowAccumulator(mg)
        fa.run_one_step()

        drainage_area = mg.at_node['drainage_area']
        drained_area = np.sum(drainage_area[mg.boundary_nodes])
        core_area = np.sum(mg.cell_area_at_node[mg.core_nodes])
        assert drained_area == core_area

# def test_passing_unnecessary_kwarg():
#     """Test that passing a bad kwarg raises a ValueError."""
#     mg = RasterModelGrid((10,10), spacing=(1, 1))
#     z = mg.add_field('topographic__elevation', mg.node_x + mg.node_y, at = 'node')
#     with pytest.raises(ValueError):
#         FlowAccumulator(mg, bad_kwarg='woo')

def test_specifying_routing_method_wrong():
    """Test specifying incorrect method for routing compatability with DepressionFinderAndRouter."""
    mg = RasterModelGrid((10,10), spacing=(1, 1))
    z = mg.add_field('topographic__elevation', mg.node_x + mg.node_y, at = 'node')

    with pytest.raises(ValueError):
        FlowAccumulator(mg, flow_director='D4',
                        depression_finder='DepressionFinderAndRouter')

    df = DepressionFinderAndRouter(mg)
    with pytest.raises(ValueError):
        FlowAccumulator(mg, flow_director='D4', depression_finder=df)


def test_field_name_array_float_case1():
    """Topography as field, runoff rate as float"""
    mg = RasterModelGrid((5,4), spacing=(1, 1))
    topographic__elevation = np.array([0.,  0.,  0., 0.,
                                       0., 21., 10., 0.,
                                       0., 31., 20., 0.,
                                       0., 32., 30., 0.,
                                       0.,  0.,  0., 0.])
    _ = mg.add_field('node', 'topographic__elevation', topographic__elevation)
    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = FlowAccumulator(mg, 'topographic__elevation', runoff_rate=10.)
    assert_array_equal(mg.at_node['water__unit_flux_in'], 10.*np.ones(mg.size('node')))

    fa.run_one_step()
    reciever = np.array([ 0,  1,  2,  3,
                          4,  1,  2,  7,
                          8, 10,  6, 11,
                         12, 14, 10, 15,
                         16, 17, 18, 19])

    da = np.array([ 0.,  1.,  5.,  0.,
                    0.,  1.,  5.,  0.,
                    0.,  1.,  4.,  0.,
                    0.,  1.,  2.,  0.,
                    0.,  0.,  0.,  0.])

    q = np.array([ 0.,  10.,  50.,  0.,
                   0.,  10.,  50.,  0.,
                   0.,  10.,  40.,  0.,
                   0.,  10.,  20.,  0.,
                   0.,   0.,   0.,  0.])

    assert_array_equal(mg.at_node['flow__receiver_node'], reciever)
    assert_array_equal(mg.at_node['drainage_area'], da)
    assert_array_equal(mg.at_node['surface_water__discharge'], q)


def test_field_name_array_float_case2():
    """Topography as field, runoff rate as field name"""
    mg = RasterModelGrid((5,4), spacing=(1, 1))
    topographic__elevation = np.array([0.,  0.,  0., 0.,
                                       0., 21., 10., 0.,
                                       0., 31., 20., 0.,
                                       0., 32., 30., 0.,
                                       0.,  0.,  0., 0.])

    runoff_rate = ([1.,  1.,  1., 1.,
                    2.,  2.,  2., 2.,
                    3.,  3.,  3., 3.,
                    4.,  4.,  4., 4.,
                    5.,  5.,  5., 5.])

    _ = mg.add_field('node', 'topographic__elevation', topographic__elevation)
    _ = mg.add_field('node', 'runoff_rate', runoff_rate)

    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = FlowAccumulator(mg, 'topographic__elevation', runoff_rate='runoff_rate')

    fa.run_one_step()
    reciever = np.array([ 0,  1,  2,  3,
                          4,  1,  2,  7,
                          8, 10,  6, 11,
                         12, 14, 10, 15,
                         16, 17, 18, 19])

    da = np.array([ 0.,  1.,  5.,  0.,
                    0.,  1.,  5.,  0.,
                    0.,  1.,  4.,  0.,
                    0.,  1.,  2.,  0.,
                    0.,  0.,  0.,  0.])

    q = np.array([  0.,   2.,  16.,   0.,   # KRB double checked these numbers by hand 5/15/18 - OK
                    0.,   2.,  16.,   0.,
                    0.,   3.,  14.,   0.,
                    0.,   4.,   8.,   0.,
                    0.,   0.,   0.,   0.])

    assert_array_equal(mg.at_node['flow__receiver_node'], reciever)
    assert_array_equal(mg.at_node['drainage_area'], da)
    assert_array_equal(mg.at_node['surface_water__discharge'], q)


def test_field_name_array_float_case3():
    """Topography as field, runoff rate as array"""
    mg = RasterModelGrid((5,4), spacing=(1, 1))
    topographic__elevation = np.array([0.,  0.,  0., 0.,
                                       0., 21., 10., 0.,
                                       0., 31., 20., 0.,
                                       0., 32., 30., 0.,
                                       0.,  0.,  0., 0.])

    runoff_rate = ([1.,  1.,  1., 1.,
                    2.,  2.,  2., 2.,
                    3.,  3.,  3., 3.,
                    4.,  4.,  4., 4.,
                    5.,  5.,  5., 5.])

    _ = mg.add_field('node', 'topographic__elevation', topographic__elevation)

    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = FlowAccumulator(mg, 'topographic__elevation', runoff_rate=runoff_rate)

    fa.run_one_step()
    reciever = np.array([ 0,  1,  2,  3,
                          4,  1,  2,  7,
                          8, 10,  6, 11,
                         12, 14, 10, 15,
                         16, 17, 18, 19])

    da = np.array([ 0.,  1.,  5.,  0.,
                    0.,  1.,  5.,  0.,
                    0.,  1.,  4.,  0.,
                    0.,  1.,  2.,  0.,
                    0.,  0.,  0.,  0.])

    q = np.array([  0.,   2.,  16.,   0.,   # KRB double checked these numbers by hand 5/15/18 - OK
                    0.,   2.,  16.,   0.,
                    0.,   3.,  14.,   0.,
                    0.,   4.,   8.,   0.,
                    0.,   0.,   0.,   0.])

    assert_array_equal(mg.at_node['flow__receiver_node'], reciever)
    assert_array_equal(mg.at_node['drainage_area'], da)
    assert_array_equal(mg.at_node['surface_water__discharge'], q)


def test_field_name_array_float_case4():
    """Topography as array, runoff rate as float"""
    mg = RasterModelGrid((5,4), spacing=(1, 1))
    topographic__elevation = np.array([0.,  0.,  0., 0.,
                                       0., 21., 10., 0.,
                                       0., 31., 20., 0.,
                                       0., 32., 30., 0.,
                                       0.,  0.,  0., 0.])
    _ = mg.add_field('node', 'topographic__elevation', topographic__elevation)
    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = FlowAccumulator(mg, topographic__elevation, runoff_rate=10.)
    assert_array_equal(mg.at_node['water__unit_flux_in'], 10.*np.ones(mg.size('node')))

    fa.run_one_step()
    reciever = np.array([ 0,  1,  2,  3,
                          4,  1,  2,  7,
                          8, 10,  6, 11,
                         12, 14, 10, 15,
                         16, 17, 18, 19])

    da = np.array([ 0.,  1.,  5.,  0.,
                    0.,  1.,  5.,  0.,
                    0.,  1.,  4.,  0.,
                    0.,  1.,  2.,  0.,
                    0.,  0.,  0.,  0.])

    q = np.array([ 0.,  10.,  50.,  0.,
                   0.,  10.,  50.,  0.,
                   0.,  10.,  40.,  0.,
                   0.,  10.,  20.,  0.,
                   0.,   0.,   0.,  0.])

    assert_array_equal(mg.at_node['flow__receiver_node'], reciever)
    assert_array_equal(mg.at_node['drainage_area'], da)
    assert_array_equal(mg.at_node['surface_water__discharge'], q)


def test_field_name_array_float_case5():
    """Topography as array, runoff rate as field name"""
    mg = RasterModelGrid((5,4), spacing=(1, 1))
    topographic__elevation = np.array([0.,  0.,  0., 0.,
                                       0., 21., 10., 0.,
                                       0., 31., 20., 0.,
                                       0., 32., 30., 0.,
                                       0.,  0.,  0., 0.])

    runoff_rate = ([1.,  1.,  1., 1.,
                    2.,  2.,  2., 2.,
                    3.,  3.,  3., 3.,
                    4.,  4.,  4., 4.,
                    5.,  5.,  5., 5.])

    _ = mg.add_field('node', 'topographic__elevation', topographic__elevation)
    _ = mg.add_field('node', 'runoff_rate', runoff_rate)

    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = FlowAccumulator(mg, 'topographic__elevation', runoff_rate='runoff_rate')

    fa.run_one_step()
    reciever = np.array([ 0,  1,  2,  3,
                          4,  1,  2,  7,
                          8, 10,  6, 11,
                         12, 14, 10, 15,
                         16, 17, 18, 19])

    da = np.array([ 0.,  1.,  5.,  0.,
                    0.,  1.,  5.,  0.,
                    0.,  1.,  4.,  0.,
                    0.,  1.,  2.,  0.,
                    0.,  0.,  0.,  0.])

    q = np.array([  0.,   2.,  16.,   0.,   # KRB double checked these numbers by hand 5/15/18 - OK
                    0.,   2.,  16.,   0.,
                    0.,   3.,  14.,   0.,
                    0.,   4.,   8.,   0.,
                    0.,   0.,   0.,   0.])

    assert_array_equal(mg.at_node['flow__receiver_node'], reciever)
    assert_array_equal(mg.at_node['drainage_area'], da)
    assert_array_equal(mg.at_node['surface_water__discharge'], q)


def test_field_name_array_float_case6():
    """Topography as array, runoff rate as array"""
    mg = RasterModelGrid((5,4), spacing=(1, 1))
    topographic__elevation = np.array([0.,  0.,  0., 0.,
                                       0., 21., 10., 0.,
                                       0., 31., 20., 0.,
                                       0., 32., 30., 0.,
                                       0.,  0.,  0., 0.])

    runoff_rate = ([1.,  1.,  1., 1.,
                    2.,  2.,  2., 2.,
                    3.,  3.,  3., 3.,
                    4.,  4.,  4., 4.,
                    5.,  5.,  5., 5.])

    _ = mg.add_field('node', 'topographic__elevation', topographic__elevation)

    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = FlowAccumulator(mg, topographic__elevation, runoff_rate=runoff_rate)

    fa.run_one_step()
    reciever = np.array([ 0,  1,  2,  3,
                          4,  1,  2,  7,
                          8, 10,  6, 11,
                         12, 14, 10, 15,
                         16, 17, 18, 19])

    da = np.array([ 0.,  1.,  5.,  0.,
                    0.,  1.,  5.,  0.,
                    0.,  1.,  4.,  0.,
                    0.,  1.,  2.,  0.,
                    0.,  0.,  0.,  0.])

    q = np.array([  0.,   2.,  16.,   0.,   # KRB double checked these numbers by hand 5/15/18 - OK
                    0.,   2.,  16.,   0.,
                    0.,   3.,  14.,   0.,
                    0.,   4.,   8.,   0.,
                    0.,   0.,   0.,   0.])

    assert_array_equal(mg.at_node['flow__receiver_node'], reciever)
    assert_array_equal(mg.at_node['drainage_area'], da)
    assert_array_equal(mg.at_node['surface_water__discharge'], q)
